import epcpm._build
import subprocess


def get_git_revision_hash() -> str:
    return subprocess.check_output(["git", "rev-parse", "HEAD"]).decode("ascii").strip()


__version__ = "2022.1.7"
__sha__ = get_git_revision_hash()
# del get_versions

__version_tag__ = "v{}".format(__version__)
__build_tag__ = epcpm._build.job_id



import pathlib

import click

import epyqlib.pm.parametermodel

import epcpm.parameterstohierarchy
import epcpm.project
import epcpm.cantosym


def relative_path(target, reference):
    reference = pathlib.Path(reference)
    if reference.is_file():
        reference = reference.parents[0]

    return pathlib.Path(target).resolve().relative_to(reference.resolve())


@click.command()
@click.option("--project", "project_file", type=click.File(), required=True)
@click.option("--sym", "sym_file", type=click.File("w"), required=True)
@click.option(
    "--hierarchy",
    "hierarchy_file",
    type=click.File("w"),
    required=True,
)
def cli(project_file, sym_file, hierarchy_file):
    project = epcpm.project.load(project_file)

    (access_levels,) = project.models.parameters.root.nodes_by_filter(
        filter=(lambda node: isinstance(node, epyqlib.pm.parametermodel.AccessLevels)),
    )

    sym_builder = epcpm.cantosym.builders.wrap(
        wrapped=project.models.can.root,
        access_levels=access_levels,
        parameter_uuid_finder=project.models.can.node_from_uuid,
        parameter_model=project.models.parameters,
    )

    hierarchy_builder = epcpm.parameterstohierarchy.builders.wrap(
        wrapped=project.models.parameters.root,
        can_root=project.models.can.root,
    )

    sym = sym_builder.gen()
    hierarchy = hierarchy_builder.gen(indent=4)
    if hierarchy[-1] != "\n":
        hierarchy += "\n"

    sym_file.write(sym)
    hierarchy_file.write(hierarchy)



import faulthandler

faulthandler.enable()

import logging
import os.path
import sys

import click
from PyQt5 import QtCore, QtGui, QtWidgets

import epyqlib.utils.qt
import epcpm.mainwindow

import epcpm.cli.utils


# See file COPYING in this source tree
__copyright__ = "Copyright 2017, EPC Power Corp."
__license__ = "GPLv2+"


def main(project, verbosity, logger):
    app = QtWidgets.QApplication(sys.argv)

    epyqlib.utils.qt.exception_message_box_register_versions(
        version_tag=epcpm.__version_tag__,
        build_tag=epcpm.__build_tag__,
    )

    sys.excepthook = epyqlib.utils.qt.exception_message_box

    os_signal_timer = epyqlib.utils.qt.setup_sigint()

    QtCore.qInstallMessageHandler(epyqlib.utils.qt.message_handler)

    app.setStyleSheet(
        "QMessageBox {{ messagebox-text-interaction-flags: {}; }}".format(
            QtCore.Qt.TextBrowserInteraction
        )
    )

    app.setOrganizationName("EPC Power Corp.")
    app.setApplicationName("EPC Parameter Management")

    if verbosity >= 1:
        logger.setLevel(logging.DEBUG)

        if verbosity >= 2:
            # twisted.internet.defer.setDebugging(True)
            pass

            if verbosity >= 3:
                logging.getLogger().setLevel(logging.DEBUG)

    window = epcpm.mainwindow.Window(
        title="EPC Parameter Manager",
        icon_path="icon.ico",
    )

    epyqlib.utils.qt.exception_message_box_register_parent(
        parent=window.main_window,
    )

    if project is not None:
        filename = os.path.abspath(project)
        window.open_project(filename=filename)

    window.main_window.show()

    return app.exec()


@click.command()
@epcpm.cli.utils.project_option()
@click.option(
    "-v",
    "--verbose",
    "verbosity",
    count=True,
    help="Increase verbosity of output (up to three times)",
)
def _entry_point(project, verbosity):
    """Parameter Manager GUI"""

    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    stream_handler = logging.StreamHandler()
    file_handler = logging.FileHandler("epcpm.log")

    for handler in (stream_handler, file_handler):
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    logging.basicConfig(format="%(asctime)s - %(levelname)s - %(message)s")

    sys.excepthook = epyqlib.utils.general.exception_logger

    return main(project=project, verbosity=verbosity, logger=logger)


# for PyInstaller
if __name__ == "__main__":
    sys.exit(_entry_point())



from pytestqt import qtbot

import epcpm.mainwindow

import epyqlib.pm.parametermodel

# See file COPYING in this source tree
__copyright__ = "Copyright 2017, EPC Power Corp."
__license__ = "GPLv2+"


def test_default_ranges():
    p = epyqlib.pm.parametermodel.Parameter(name="default_min_max")
    assert p.minimum is None
    assert p.maximum is None


# tests/test_parameters.py::test_gui_launch ERROR: InvocationError: '/epc/t/472/p/.tox/test/bin/python -m pytest -v'
# def test_gui_launch(qtbot):
#     window = pm.mainwindow.Window(ui_file='__main__.ui')
#     window.show()
#
#     qtbot.addWidget(window)



import json
import pathlib

import click
import graham

import epyqlib.pm.valuesetmodel

import epcpm.project
import epcpm.symtoproject


@click.command()
@click.option("--sym", type=click.File("rb"), required=True)
@click.option("--hierarchy", type=click.File(), required=True)
@click.option("--parameters", type=click.File())
@click.option("--value-set", type=click.File("w"), required=True)
def cli(sym, hierarchy, parameters, value_set):
    value_set_file = value_set

    parameters_root, can_root = epcpm.symtoproject.load_can_file(
        can_file=sym,
        file_type=str(pathlib.Path(sym.name).suffix[1:]),
        parameter_hierarchy_file=hierarchy,
    )

    project_model = epcpm.project.Project(
        models=epcpm.project.Models(
            parameters=epyqlib.attrsmodel.Model(
                root=parameters_root,
                columns=epyqlib.pm.parametermodel.columns,
            ),
            can=epyqlib.attrsmodel.Model(
                root=can_root,
                columns=epcpm.canmodel.columns,
            ),
        ),
    )

    value_set = epyqlib.pm.valuesetmodel.create_blank(
        parameter_model=project_model.models.parameters,
    )

    (parameters_root,) = (
        node
        for node in value_set.parameter_model.root.children
        if node.name == "Parameters"
    )

    epyqlib.pm.valuesetmodel.copy_parameter_data(
        value_set=value_set,
        human_names=False,
        base_node=parameters_root,
        calculate_unspecified_min_max=True,
        can_root=can_root,
    )

    value_set_parameters = {
        parameter.name: parameter for parameter in value_set.model.root.children
    }

    parameters = json.load(parameters)

    for name, value in parameters.items():
        value_set_parameters[name].value = value

    value_set_file.write(graham.dumps(value_set.model.root, indent=4).data)



import decimal
import itertools
import os
import pathlib
import re

import canmatrix
import click.testing
import pytest

import epyqlib.tests.common

import epcpm.cli.importsym
import epcpm.cli.exportsym
import epcpm.importexport
import epcpm.importexportdialog


round_trip = pathlib.Path(__file__).parents[3] / "roundtrip"
project = round_trip / "project.pmp"
example_device = epyqlib.tests.common.new_devices()[("develop", "factory")]
exported_can = (round_trip / "exported").with_suffix(example_device.can.suffix)
exported_hierarchy = (round_trip / "hierarchy.parameters.json").with_suffix(
    example_device.hierarchy.suffix
)


def _full_round_trip():
    # directory = pathlib.Path('full_import')
    # paths = epcpm.importexportdialog.paths_from_directory(directory)
    # project = epcpm.importexport.full_import(paths=paths)
    import epcpm.project

    project = epcpm.project.loadp(pathlib.Path("full_import") / "project.pmp")

    directory = pathlib.Path(os.sep) / "epc" / "g" / "36" / "grid-tied"
    paths = epcpm.importexportdialog.paths_from_directory(directory)

    epcpm.importexport.full_export(project=project, paths=paths, first_time=True)

    # project = epcpm.importexport.full_import(paths=paths)
    # epcpm.importexport.full_export(project=project, paths=paths)


@pytest.mark.skip
def test_import():
    runner = click.testing.CliRunner()

    result = runner.invoke(
        epcpm.cli.importsym.cli,
        [
            "--sym",
            example_device.can,
            "--hierarchy",
            example_device.hierarchy,
            "--project",
            project,
            "--can",
            round_trip / "can.json",
            "--parameters",
            round_trip / "parameters.json",
            "--sunspec",
            round_trip / "sunspec.json",
            "--add-tables",
            "--add-sunspec-types",
        ],
        catch_exceptions=False,
    )
    print(result.output)

    assert result.exit_code == 0


@pytest.mark.skip
def test_export():
    runner = click.testing.CliRunner()

    result = runner.invoke(
        epcpm.cli.exportsym.cli,
        [
            "--project",
            project,
            "--sym",
            exported_can,
            "--hierarchy",
            exported_hierarchy,
        ],
        catch_exceptions=False,
    )
    print(result.output)

    assert result.exit_code == 0


def normalized_attributes(attributes):
    attributes = dict(attributes)

    return attributes


def assert_signals_equal(
    original,
    original_frame,
    exported,
    exported_frame,
):
    print(
        original_frame.name,
        original_frame.mux_names.get(original.multiplex),
        original.name,
    )
    assert_attributes_equal(
        original=original,
        exported=exported,
        attributes=(
            "enumeration",
            "is_float",
            "is_little_endian",
            "is_multiplexer",
            "is_signed",
            "max",
            "min",
            "multiplex",
            "factor",
            "offset",
            "size",
            "startBit",
            "unit",
        ),
    )

    tweaked = dict(exported.attributes)
    if "LongName" not in original.attributes:
        tweaked.pop("LongName", None)

    tweaked_attributes = normalized_attributes(tweaked)
    original_attributes = normalized_attributes(original.attributes)

    assert tweaked_attributes == original_attributes, original.name

    factory = "<factory>"
    configs = ("<HY>", "<MG3>", "<MG4>", "<DG>", "<DC>")

    def split_and_access_level(comment):

        split = comment.split()

        return (
            factory in split,
            (tag in split for tag in configs),
            [x for x in split if x not in (factory, *configs)],
        )

    original_factory, original_configs, original_comment = split_and_access_level(
        original.comment
    )
    exported_factory, exported_configs, exported_comment = split_and_access_level(
        exported.comment
    )

    assert exported_comment == original_comment, original.name

    not_applicable = (
        "readparam_command",
        "readparam_status",
        "meta",
    )
    if original.name.casefold() not in not_applicable:
        original_mux = [
            x for x in original_frame.signals if x.multiplex == "Multiplexor"
        ]
        (original_mux,) = original_mux if original_mux else (None,)

        exported_mux = [
            x for x in exported_frame.signals if x.multiplex == "Multiplexor"
        ]
        (exported_mux,) = exported_mux if exported_mux else (None,)

        if original_mux is None:
            assert exported_mux == original_mux, original.name

        def tagged_including_mux_check(tag, tagged, signal, mux):
            if mux is not None:
                return tagged or (
                    signal is not mux and tag in mux.comments[signal.multiplex]
                )

            return tagged or tag in original_frame.comment

        original_access_level = tagged_including_mux_check(
            tag=factory,
            tagged=original_factory,
            signal=original,
            mux=original_mux,
        )
        exported_access_level = tagged_including_mux_check(
            tag=factory,
            tagged=exported_factory,
            signal=exported,
            mux=exported_mux,
        )

        assert exported_access_level == original_access_level, original.name

        z = zip(configs, original_configs, exported_configs)
        for config, original_tagged, exported_tagged in z:
            original_config = tagged_including_mux_check(
                tag=config,
                tagged=original_tagged,
                signal=original,
                mux=original_mux,
            )
            exported_config = tagged_including_mux_check(
                tag=config,
                tagged=exported_tagged,
                signal=exported,
                mux=exported_mux,
            )

            assert exported_config == original_config, (
                config,
                exported.comment,
                original.comment,
                original.name,
            )

    assert_varied_dict_equal(original.values, exported.values)

    assert exported_comment == original_comment, original.name


def get_name(x):
    return x.name


def sorted_names(names):
    return sorted(names, key=lambda x: (x.casefold(), x))


def unvary_signal_names(name):
    changes = (
        (r"(IGBT|Diode)_R(\d)", r"\1_r\2"),
        (r"(IGBT|Diode)_Tc(\d)", r"\1_tc\2"),
        ("Rii", "rii"),
        ("Rdi", "rdi"),
        ("Alpha", "alpha"),
        ("Amps", "amps"),
        ("Beta", "beta"),
        (r"Leg(\d)", r"leg\1"),
        ("CurrentLCL", "currentLCL"),
        ("MinPulseTime", "minPulseTime"),
        ("N15V_Supply", "n15V_Supply"),
        (r"V([p\d]+)_Supply", r"v\1_Supply"),
        ("Phase([ABC])", r"phase\1"),
        ("PostTriggerDuration_S", "PostTriggerDuration_s"),
        (r"LegInductance(\d)", r"legInductance\1"),
        ("TempSetpoint", "tempSetpoint"),
        ("LineCurrentLimit", "lineCurrentLimit"),
        ("Dc(Current|Voltage)Limit", r"dc\1Limit"),
        ("Dc(Current|Voltage)Limit_Echo", r"dc\1Limit_echo"),
        ("Line(Current|Voltage)Limit_Echo", r"line\1Limit_echo"),
        ("PpcIslDroop([VF])_IBpt", r"PpcIslDroop\1_iBpt"),
        ("PpcIslDroop([VF])_DBpt", r"PpcIslDroop\1_dBpt"),
        ("PpcIslDroop([VF])_QBpt", r"PpcIslDroop\1_qBpt"),
        ("PpcIslDestab([VF])_VBpt", r"PpcIslDestab\1_vBpt"),
        ("PpcIslDestab([VF])_DBpt", r"PpcIslDestab\1_dBpt"),
        ("PpcIslDestab([VF])_QBpt", r"PpcIslDestab\1_qBpt"),
        ("PpcIslDestab([VF])_FBpt", r"PpcIslDestab\1_fBpt"),
        ("Ig_Max", "Ig_max"),
        (
            r"(MesdConfigLutCalibration_)T(Igbt|Aux|Motor)_(\d+)",
            r"\1t\2_\3",
        ),
        (
            "(ProtTemperatureLimitTrip_)T({})_(Running|Stopped)".format(
                "|".join(
                    (
                        "Inverter",
                        "Internal",
                        "Device",
                        "Inlet",
                        "Delta",
                    )
                ),
            ),
            lambda m: f"{m.group(1)}t{m.group(2)}_{m.group(3).casefold()}",
        ),
        (
            "(ProtCtrlVoltageLimitTrip)_V(Pos|Neg)(24|15|5|3)",
            r"\1_v\2\3",
        ),
        (r"(.*)_TxRate", r"\1_txRate"),
        *(
            (x, lambda m: m.group(0)[0].casefold() + m.group(0)[1:])
            for x in (
                "DIg",
                "DpwmStrategy",
                "DcGroup",
                "D[012]",
                "F[012]",
                "Q([012]|Ig)",
                "V([012])",
                "(F|V|W|XpVa)Tau",
                "AcGroup",
                "(C|G)(L|R)s",
                "AcHalfPeriodTrack",
                "DtCompEnable",
                "DutyLimit",
                "EeSaveInProgress",
                "EntryPulseNr",
                "FCut(F|Ig|V|Vg)",
                "FDev(H|L)",
                "FMax",
                "FMin",
                "FNom",
                "FourWireEnable",
                "FSw(Hysteresis|Minimum|Request)",
                "I([012]|Abs|Delta|Dlt)",
                "Ind(IslStart|Threshold)",
                "K(Alpha|Apply|Phi|Vq)",
                "Limit(XY|Z)",
                "Lim(Lower|Upper)",
                "Load(L|PhiTau|R)",
                "Min(Current|VectorTime|Volt)",
                "ModDepthLimit(Beg|End)",
                "NetworkFrequency",
                "P(Delta|PhaseSel)",
                "PhaseSel",
                "PhiAbs",
                "Sat(CompEnable|Voltage)",
                "S(C|R)s",
                "SRv",
                "SimModeEnable",
                "SpectrumAdjustTau",
                "SpreadSpect(Enable|Inject)",
                "StandardDeviation",
                "SyncMod(Enable|Strategy)",
                "Tau",
                "T(Adapt|Deadband|Delta)",
                "Threshold",
                "TimeConst",
                "TrigCalTime",
                "TripTime",
                "TSwitch",
                "U",
                "VDelta",
                "VectAlignCtrl_(K[dp]|StepSat|Ti)",
                "V(Max|Min)",
                "VoltMag",
                "VsComp(Enable|Tau)",
                "Y",
                "ZeroVectorStrategy",
            )
        ),
        *(
            (
                r"(.*)_({x})_(.*)".format(x=x),
                lambda m: f"{m.group(1)}_{m.group(2).casefold()}_{m.group(3)}",
            )
            for x in (
                "Seconds",
                "Hertz",
                "Percent_Nominal_Volts",
                "Percent_Nominal_Var",
                "Percent_Nominal_Pwr",
                "Percent",
                "Nominal_Volts",
            )
        ),
        *(
            (x, lambda m: m.group(1) + m.group(2) + m.group(3).casefold())
            for x in (
                "(DcCurrent|DcVoltage|ReactivePower|RealPower)()(Limit|_Echo)",
                "(ReadParam|SaveToEE)(_)(Command|Status)",
                "(TempTmodel)(_)(Calculated)",
                "({})(_)(Echo)".format(
                    "|".join(
                        (
                            "Enable",
                            "FaultClr",
                            "WarningClr",
                            "ReactiveCurrent",
                            "RealCurrent",
                            "IslandReconnect",
                        )
                    ),
                ),
                "({})(_)(Measured)".format(
                    "|".join(
                        (
                            "TempDeviceBrdg[ABC]",
                            "CurrentAC",
                            "Frequency",
                            "VoltageAC",
                            "FrequencyRemote",
                            "VoltageAcRemote",
                            "SolarCurrent",
                            "SolarVoltage",
                            "CurrentDC",
                            "VoltageDCBus",
                            "VoltageDCInput",
                            "L[123]Current",
                            "L[123]ToL[123]Voltage",
                            ".*",
                        )
                    ),
                ),
            )
        ),
    )

    tweaked = name
    for pattern, replacement in changes:
        tweaked = re.sub(f"^{pattern}$", replacement, tweaked)

    return tweaked


def assert_signal_list_equal(
    original,
    original_frame,
    exported,
    exported_frame,
):
    original_names = sorted_names(x.name for x in original)
    exported_names = sorted_names(unvary_signal_names(x.name) for x in exported)

    case_insensitive = (
        "aclosshotstandbydc",
        "aclosshotstandbypc",
        "acwaitenabledcatref",
        "argadjust",
        "argerrmax",
        "argerrsat",
        "argerrtau",
        "autodisconnectongridloss",
        "autofollowingpostconnect",
        "autoreconnectongridavail",
        "autosequencewhendisabled",
        "autovolttrackduringsynch",
        "conntimeout",
        "controlsvariant",
        "current",
        "dcwaitenabledcatref",
        "disctimeout",
        "enabledelayac",
        "enabledelaydc",
        "didt",
        "dqccterm",
        "dummy",
        "freq",
        "fswactual",
        # 'ftau',
        "gain",
        "groupedenable",
        "hardwarevariant",
        "level",
        "limit",
        "magadjust",
        "magerrmax",
        "magerrtau",
        "moddepthxy",
        "moddepthz",
        "max",
        "min",
        "pllenable",
        "pulsenractual",
        "r",
        "rate",
        "resistance",
        "syncactive",
        "syncholdtime",
        "synctimeout",
        "tc",
        "temp",
        "temperature",
        "time",
        "tisllatch",
        "tunload",
        "vdevh",
        "vdevl",
        "volts",
        # 'vtau',
        # 'wtau',
    )

    zipped = itertools.zip_longest(original_names, exported_names)
    for original_name, exported_name in zipped:
        if original_name.casefold() in case_insensitive:
            assert exported_name.casefold() == original_name.casefold(), original.name
        else:
            assert exported_name == original_name

    def key(x):
        return (
            x.name.casefold(),
            x.multiplex if x.multiplex is not None else float("-inf"),
        )

    zipped = itertools.zip_longest(
        sorted(original, key=key),
        sorted(exported, key=key),
    )
    for original_signal, exported_signal in zipped:
        assert_signals_equal(
            original=original_signal,
            original_frame=original_frame,
            exported=exported_signal,
            exported_frame=exported_frame,
        )


def assert_attributes_equal(original, exported, *, attributes):
    for attribute in attributes:
        assert getattr(exported, attribute) == getattr(original, attribute), (
            original.name,
            attribute,
        )


def assert_varied_dict_equal(original, exported):
    zipped = itertools.zip_longest(
        sorted(original.items()),
        sorted(exported.items()),
    )
    for (original_value, original_name), (exported_value, exported_name) in zipped:
        assert exported_value == original_value

        tweaked = unvary_signal_names(exported_name)
        assert tweaked == original_name


def assert_frames_equal(original, exported):
    assert_attributes_equal(
        original=original,
        exported=exported,
        attributes=(
            "name",
            "id",
            "extended",
            "attributes",
            "comment",
            "is_multiplexed",
            "size",
        ),
    )

    assert_varied_dict_equal(original.mux_names, exported.mux_names)

    assert_signal_list_equal(
        original=original.signals,
        original_frame=original,
        exported=exported.signals,
        exported_frame=exported,
    )


def collect_attribute(sequence, name):
    return {getattr(x, name) for x in sequence}


def assert_frame_lists_equal(original, exported):
    original_names = collect_attribute(original, "name")
    exported_names = collect_attribute(exported, "name")

    assert exported_names == original_names, original.name

    zipped = itertools.zip_longest(
        sorted(original, key=get_name),
        sorted(exported, key=get_name),
    )
    for original_frame, exported_frame in zipped:
        assert_frames_equal(original_frame, exported_frame)


def assert_value_tables_equal(original, exported):
    return exported == original


@pytest.mark.skip
def test_roundtrip():
    (original,) = canmatrix.formats.loadp(os.fspath(example_device.can)).values()
    assert original.load_errors == []
    (exported,) = canmatrix.formats.loadp(os.fspath(exported_can)).values()
    assert exported.load_errors == []

    assert_frame_lists_equal(original.frames, exported.frames)
    assert_value_tables_equal(original.value_tables, exported.value_tables)



# This file has been generated


build_system = 'GitHub Actions'
build_id = '1780300523'
build_number = '259'
build_version = None
job_id = '1780300523_259'
job_url = 'https://github.com/epcpower/pm/actions/runs/1780300523'



import pytest

import epcpm.symtoproject
import epyqlib.pm


@pytest.fixture
def access_levels():
    access_levels = epyqlib.pm.parametermodel.AccessLevels()
    user_level = epyqlib.pm.parametermodel.AccessLevel(name="User", value=0)
    factory_level = epyqlib.pm.parametermodel.AccessLevel(name="Factory", value=1)

    access_levels.append_child(user_level)
    access_levels.append_child(factory_level)

    return access_levels


@pytest.fixture
def variant_cfgs():
    variants = epyqlib.pm.parametermodel.Enumeration()
    enumerator_type = epyqlib.pm.parametermodel.Enumerator

    cfgs = ["None", "MG3", "MG4", "DG", "HY", "DC"]
    val = 0
    for cfg in cfgs:
        e = enumerator_type(name=cfg, value=val)
        variants.append_child(e)
        val = val + 1

    variants = [variant for variant in variants.children if variant.name != "None"]

    return variants


def test_strip_access_level_factory(access_levels):
    comment = "abc <factory> def"

    stripped, level = epcpm.symtoproject.strip_access_level(comment, access_levels)

    assert stripped.split() == ["abc", "def"]
    assert level == access_levels.by_name("factory")


def test_strip_access_level_normal(access_levels):
    comment = "abc def"

    stripped, level = epcpm.symtoproject.strip_access_level(comment, access_levels)

    assert stripped == comment
    assert level == access_levels.by_name("user")


def test_strip_variant_parameter_specific(variant_cfgs):
    comment = "abc <HY> <DG> <MG4> def"

    stripped, variants = epcpm.symtoproject.strip_variant_parameter_tag(
        comment, variant_cfgs
    )
    variant_names = [variant.name for variant in variants]
    assert stripped.split() == ["abc", "def"]
    assert variant_names == ["MG4", "DG", "HY"]


def test_strip_variant_parameter_normal(variant_cfgs):
    comment = "abc def"

    stripped, variants = epcpm.symtoproject.strip_variant_parameter_tag(
        comment, variant_cfgs
    )
    variant_names = [variant.name for variant in variants]
    assert stripped == comment
    assert variant_names == ["MG3", "MG4", "DG", "HY", "DC"]



"""
Testing for parameterstoc script cli.

It is possible that the parameterstoc module is either out of date
or no longer applicable for normal use.
"""
import pathlib
import textwrap

import click
import click.testing

import epcpm.cli.parameterstoc


parameters_path = (
    pathlib.Path(__file__).parents[0] / "test_parameterstoc_parameters.json"
)


def test_declaration():
    runner = click.testing.CliRunner()
    result = runner.invoke(
        epcpm.cli.parameterstoc.cli,
        [
            "--parameters",
            parameters_path,
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert result.output == textwrap.dedent(
        """\
    struct GreenType_s
    {
      RedType red;
    };
    typedef struct GreenType_s
    {
    } GreenType_t;
    """
    )


def test_instantiation():
    runner = click.testing.CliRunner()
    result = runner.invoke(
        epcpm.cli.parameterstoc.cli,
        [
            "--parameters",
            parameters_path,
            "--instantiation",
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert result.output == textwrap.dedent(
        """\
    GreenType_t green;
    BlueType blue;
    """
    )



import pathlib

import attr
import graham
import marshmallow

import epyqlib.pm.parametermodel
import epyqlib.utils.qt

import epcpm.canmodel
import epcpm.sunspecmodel
import epcpm.staticmodbusmodel


class ProjectSaveCanceled(Exception):
    pass


# class ProjectLoadCanceled(Exception):
#     pass


def create_blank():
    project = Project()
    _post_load(project)

    return project


def loads(s, project_path=None, post_load=True):
    project = graham.schema(Project).loads(s).data

    if project_path is not None:
        project.filename = pathlib.Path(project_path).absolute()

    if post_load:
        _post_load(project)

    return project


def load(f, post_load=True):
    project = loads(f.read(), project_path=f.name, post_load=post_load)

    return project


def loadp(path, post_load=True):
    with open(path) as f:
        return load(f, post_load=post_load)


def _post_load(project):
    models = project.models

    if models.parameters is None:
        if project.paths.parameters is None:
            models.parameters = epyqlib.attrsmodel.Model(
                root=epyqlib.pm.parametermodel.Root(),
                columns=epyqlib.pm.parametermodel.columns,
            )
        else:
            models.parameters = load_model(
                project=project,
                path=project.paths.parameters,
                root_type=epyqlib.pm.parametermodel.Root,
                columns=epyqlib.pm.parametermodel.columns,
            )

    if models.can is None:
        if project.paths.can is None:
            models.can = epyqlib.attrsmodel.Model(
                root=epcpm.canmodel.Root(),
                columns=epcpm.canmodel.columns,
            )
        else:
            models.can = load_model(
                project=project,
                path=project.paths.can,
                root_type=epcpm.canmodel.Root,
                columns=epcpm.canmodel.columns,
            )

    if models.sunspec is None:
        if project.paths.sunspec is None:
            models.sunspec = epyqlib.attrsmodel.Model(
                root=epcpm.sunspecmodel.Root(),
                columns=epcpm.sunspecmodel.columns,
            )
        else:
            models.sunspec = load_model(
                project=project,
                path=project.paths.sunspec,
                root_type=epcpm.sunspecmodel.Root,
                columns=epcpm.sunspecmodel.columns,
                drop_sources=(models.parameters,),
            )

    if models.staticmodbus is None:
        if project.paths.staticmodbus is None or len(project.paths.staticmodbus) == 0:
            models.staticmodbus = epyqlib.attrsmodel.Model(
                root=epcpm.staticmodbusmodel.Root(),
                columns=epcpm.staticmodbusmodel.columns,
                drop_sources=(models.parameters,),
            )
        else:
            models.staticmodbus = load_model(
                project=project,
                path=project.paths.staticmodbus,
                root_type=epcpm.staticmodbusmodel.Root,
                columns=epcpm.staticmodbusmodel.columns,
                drop_sources=(models.parameters,),
            )

    models.parameters.droppable_from.add(models.parameters)

    models.can.droppable_from.add(models.parameters)
    models.can.droppable_from.add(models.can)

    models.sunspec.droppable_from.add(models.parameters)
    models.sunspec.droppable_from.add(models.sunspec)

    models.update_enumeration_roots()


@graham.schemify(tag="models")
@attr.s
class Models:
    parameters = attr.ib(
        default=None,
        metadata=graham.create_metadata(field=marshmallow.fields.String()),
    )
    can = attr.ib(
        default=None,
        metadata=graham.create_metadata(field=marshmallow.fields.String()),
    )
    sunspec = attr.ib(
        default=None,
        metadata=graham.create_metadata(field=marshmallow.fields.String()),
    )
    staticmodbus = attr.ib(
        default=None,
        metadata=graham.create_metadata(field=marshmallow.fields.String()),
    )

    @classmethod
    def __iter__(cls):
        return (field.name for field in attr.fields(cls))

    def set_all(self, value):
        self.parameters = value
        self.can = value
        self.sunspec = value
        self.staticmodbus = value

    def items(self):
        return attr.asdict(self, recurse=False).items()

    def values(self):
        return attr.asdict(self, recurse=False).values()

    def __getitem__(self, item):
        if isinstance(item, str):
            return attr.asdict(self, recurse=False)[item]

        return attr.astuple(self, recurse=False)[item]

    def __setitem__(self, item, value):
        if isinstance(item, str):
            setattr(self, item, value)
            return

        setattr(self, attr.fields(type(self))[item].name, value)

    def update_enumeration_roots(self):
        enumerations_root = [
            child
            for child in self.parameters.root.children
            if child.name == "Enumerations"
        ]
        if len(enumerations_root) == 0:
            enumerations_root = None
        else:
            (enumerations_root,) = enumerations_root
        self.parameters.list_selection_roots["enumerations"] = enumerations_root

        if enumerations_root is None:
            access_level_root = None
        else:
            (access_level_root,) = (
                child
                for child in enumerations_root.children
                if child.name == "AccessLevel"
            )
        self.parameters.list_selection_roots["access level"] = access_level_root

        if enumerations_root is None:
            visibility_root = None
        else:
            (visibility_root,) = (
                child
                for child in enumerations_root.children
                if child.name == "CmmControlsVariant"
            )
        self.parameters.list_selection_roots["visibility"] = visibility_root

        if enumerations_root is None:
            sunspec_types_root = None
            staticmodbus_types_root = None
        else:
            sunspec_types_root = [
                child
                for child in enumerations_root.children
                if child.name == "SunSpecTypes"
            ]
            if len(sunspec_types_root) == 1:
                (sunspec_types_root,) = sunspec_types_root
            else:
                sunspec_types_root = None

            staticmodbus_types_root = [
                child
                for child in enumerations_root.children
                if child.name == "StaticModbusTypes"
            ]
            if len(staticmodbus_types_root) == 1:
                (staticmodbus_types_root,) = staticmodbus_types_root
            else:
                staticmodbus_types_root = None

        self.parameters.list_selection_roots["sunspec types"] = sunspec_types_root
        self.sunspec.list_selection_roots["sunspec types"] = sunspec_types_root
        self.sunspec.list_selection_roots["enumerations"] = enumerations_root

        self.staticmodbus.list_selection_roots[
            "staticmodbus types"
        ] = staticmodbus_types_root
        self.staticmodbus.list_selection_roots["enumerations"] = enumerations_root

        self.can.list_selection_roots["enumerations"] = enumerations_root
        self.parameters.update_nodes()
        self.can.update_nodes()
        self.sunspec.update_nodes()
        self.staticmodbus.update_nodes()


@graham.schemify(tag="project")
@attr.s
class Project:
    paths = attr.ib(
        default=attr.Factory(Models),
        metadata=graham.create_metadata(
            field=marshmallow.fields.Nested(graham.schema(Models)),
        ),
    )
    filename = attr.ib(default=None)
    models = attr.ib(default=attr.Factory(Models))
    filters = attr.ib(default=(("Parameter Project", ["pmp"]), ("All Files", ["*"])))
    data_filters = attr.ib(default=(("Dataset", ["json"]), ("All Files", ["*"])))

    def save(self, parent=None):
        if self.filename is None:
            project_path = epyqlib.utils.qt.file_dialog(
                filters=self.filters,
                parent=parent,
                save=True,
                caption="Save Project As",
            )

            if project_path is None:
                raise ProjectSaveCanceled()

            self.filename = pathlib.Path(project_path)

        project_directory = self.filename.parents[0]

        paths = Models()

        for name, path in self.paths.items():
            if path is not None:
                paths[name] = path
            else:
                new_path = epyqlib.utils.qt.file_dialog(
                    filters=self.data_filters,
                    parent=parent,
                    save=True,
                    caption="Save {} As".format(name.title()),
                )

                if new_path is None:
                    raise ProjectSaveCanceled()

                paths[name] = pathlib.Path(new_path).relative_to(project_directory)

        self.paths = paths

        with open(self.filename, "w", newline="\n") as f:
            s = graham.dumps(self, indent=4).data
            f.write(s)

            if not s.endswith("\n"):
                f.write("\n")

        for path, model in zip(paths.values(), self.models.values()):
            s = graham.dumps(model.root, indent=4).data

            with open(project_directory / path, "w", newline="\n") as f:
                f.write(s)

                if not s.endswith("\n"):
                    f.write("\n")


def load_model(project, path, root_type, columns, drop_sources=()):
    resolved_path = path
    if project.filename is not None:
        resolved_path = project.filename.parents[0] / resolved_path

    with open(resolved_path) as f:
        raw = f.read()

    root_schema = graham.schema(root_type)
    root = root_schema.loads(raw).data

    def collect(node, payload):
        payload[node.uuid] = node

    uuid_to_node = {}
    root.traverse(call_this=collect, payload=uuid_to_node, internal_nodes=True)

    def update(node, payload):
        for field in attr.fields(type(node)):
            if field.metadata.get(graham.core.metadata_key) is None or not isinstance(
                field.metadata.get(graham.core.metadata_key).field,
                epyqlib.attrsmodel.Reference,
            ):
                continue

            value = getattr(node, field.name)
            original = payload.get(value)
            if original is not None:
                setattr(node, field.name, original)

    root.traverse(call_this=update, payload=uuid_to_node, internal_nodes=True)

    return epyqlib.attrsmodel.Model(
        root=root,
        columns=columns,
        drop_sources=drop_sources,
    )



import click


def project_option(required=False):
    return click.option(
        "--project",
        type=click.Path(exists=True, dir_okay=False),
        required=required,
        help=".pmp file to load",
    )


def target_path_option(required):
    return click.option(
        "--target-path",
        type=click.Path(exists=True, file_okay=False, resolve_path=True),
        required=required,
        help="Path to the embedded project to operate on",
    )



# This file was automatically generated by sub\epyqlib\epyqlib\cli\generateversion.py
__version_build_time__ = '2022-02-01T19:34:15.351056'

__version__ = '0.0.0'
__sha__ = '968bf51d91800817193a04d0bb5c5354e7fcc1d6'
__revision__ = '968bf51d91800817193a04d0bb5c5354e7fcc1d6'



import itertools

import attr
import graham
import marshmallow

import epyqlib.attrsmodel
import epyqlib.checkresultmodel
import epyqlib.pm.parametermodel
import epyqlib.utils
import epyqlib.utils.qt
from PyQt5 import QtCore
from PyQt5 import QtWidgets


class ConsistencyError(Exception):
    pass


class MismatchedSizeAndTypeError(Exception):
    pass


class TypeNotFoundError(Exception):
    pass


def build_staticmodbus_types_enumeration():
    enumeration = epyqlib.pm.parametermodel.Enumeration(
        name="StaticModbusTypes",
        uuid="5a768d49-565d-4ffd-9c4d-f937d29f18bf",
    )

    enumerators = [
        epyqlib.pm.parametermodel.Enumerator(
            name="int16", value=1, uuid="ead5f606-8846-4dfb-bd30-d50100f29389"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="uint16", value=1, uuid="ff23a077-3d9c-4dc2-be8a-51a077058d14"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="int32", value=2, uuid="f93514e2-f173-46f9-a8c2-1a4cd15c6904"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="uint32", value=2, uuid="5ef749cd-ea55-44c9-85d7-1d574c350a84"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="staticmodbussf", value=1, uuid="2b2c843b-4f80-4822-a806-1e3cc4342ee3"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="enum16", value=1, uuid="45e7fe5e-dcb5-4a43-8455-861be6b45cbc"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="bitfield16", value=1, uuid="7d02a033-0295-417f-bc5c-838e798f938d"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="bitfield32", value=2, uuid="47755118-41dc-48ab-8e20-00f9f80d1096"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="string", value=0, uuid="fa216e96-9fea-4240-a876-1d1ed7d67c0c"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="acc16", value=1, uuid="952b419d-828d-4949-8e08-8cbd5fee6b62"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="acc32", value=2, uuid="2a948a8d-e766-4ab1-88b9-16f7c32cfe9e"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="acc64", value=4, uuid="1c98382d-11b6-4ee8-9447-3834f6f103fb"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="count", value=1, uuid="0c40dc3d-77be-481d-a3f4-784a89e0cf84"
        ),
        epyqlib.pm.parametermodel.Enumerator(
            name="pad", value=1, uuid="ed0c4a1e-777a-4423-bccc-3a4ac0a3f5be"
        ),
    ]

    for enumerator in enumerators:
        enumeration.append_child(enumerator)

    return enumeration


def create_size_attribute(default=0):
    return attr.ib(
        default=default,
        converter=int,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Integer(),
        ),
    )


def create_factor_uuid_attribute():
    return epyqlib.attrsmodel.attr_uuid(
        default=None,
        human_name="Scale Factor",
        allow_none=True,
        data_display=name_from_uuid,
        list_selection_path=("/"),
        # list_selection_path=("..", "..", "Fixed Block"),
        override_delegate=ScaleFactorDelegate,
    )


def create_parameter_uuid_attribute():
    return epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
        human_name="Parameter",
        data_display=name_from_uuid_and_parent,
        editable=False,
    )


# TODO: CAMPid 8695426542167924656654271657917491654
def name_from_uuid(node, value, model):
    if value is None:
        return None

    try:
        target_node = model.node_from_uuid(value)
    except epyqlib.attrsmodel.NotFoundError:
        return str(value)

    # return model.node_from_uuid(target_node.parameter_uuid).abbreviation


# TODO: CAMPid 8695426542167924656654271657917491654
def name_from_uuid_and_parent(node, value, model):
    if value is None:
        return None

    try:
        target_node = model.node_from_uuid(value)
    except epyqlib.attrsmodel.NotFoundError:
        return str(value)

    return "{} - {}".format(target_node.tree_parent.name, target_node.name)


class ScaleFactorDelegate(QtWidgets.QStyledItemDelegate):
    def __init__(self, text_column_name, root, parent):
        super().__init__(parent)

        self.root = root

    def createEditor(self, parent, option, index):
        return QtWidgets.QListWidget(parent=parent)

    def setEditorData(self, editor, index):
        model_index = epyqlib.attrsmodel.to_source_model(index)
        model = model_index.model()

        item = model.itemFromIndex(model_index)
        attrs_model = item.data(epyqlib.utils.qt.UserRoles.attrs_model)

        raw = model.data(model_index, epyqlib.utils.qt.UserRoles.raw)

        points = []
        for pt in self.root.children:
            type_node = attrs_model.node_from_uuid(pt.type_uuid)
            if type_node.name == "sunssf":
                points.append(pt)

        it = QtWidgets.QListWidgetItem(editor)
        it.setText("")
        it.setData(epyqlib.utils.qt.UserRoles.raw, "")
        it.setSelected(True)
        for p in points:
            it = QtWidgets.QListWidgetItem(editor)
            param = attrs_model.node_from_uuid(p.parameter_uuid)
            # it.setText(param.abbreviation)
            it.setData(epyqlib.utils.qt.UserRoles.raw, p.uuid)
            if p.uuid == raw:
                it.setSelected(True)

        editor.setMinimumHeight(editor.sizeHint().height())
        editor.itemClicked.connect(
            lambda: epyqlib.attrsmodel.hide_popup(editor),
        )
        editor.show()

    def setModelData(self, editor, model, index):
        selected_item = editor.currentItem()
        datum = str(selected_item.data(epyqlib.utils.qt.UserRoles.raw))
        model.setData(index, datum)


@graham.schemify(tag="function_data", register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class FunctionData(epyqlib.treenode.TreeNode):
    factor_uuid = create_factor_uuid_attribute()
    parameter_uuid = create_parameter_uuid_attribute()

    # hand_coded_getter = epyqlib.attrsmodel.create_checkbox_attribute(
    #     default=False,
    # )
    # hand_coded_setter = epyqlib.attrsmodel.create_checkbox_attribute(
    #     default=False,
    # )

    not_implemented = epyqlib.attrsmodel.create_checkbox_attribute(
        default=False,
    )

    type_uuid = epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
        human_name="Type",
        data_display=epyqlib.attrsmodel.name_from_uuid,
        list_selection_root="staticmodbus types",
    )

    # offset = attr.ib(
    #     default=0,
    #     converter=int,
    # )  # this is somewhat redundant with the position in the list of data
    # # points but helpful for keeping things from incidentally floating
    # # around especially in custom models where we have no sunspec
    # # model to be validating against
    # # for now, yes, this is vaguely nondescript of address vs block offset
    # @QtCore.pyqtProperty("PyQt_PyObject")
    # def pyqtify_offset(self):
    #     block_offset = getattr(self.tree_parent, "block_offset", None)
    #
    #     if block_offset is None:
    #         return None
    #
    #     return block_offset + self.block_offset
    #
    # # TODO: shouldn't this be read only?
    # @pyqtify_offset.setter
    # def pyqtify_offset(self, value):
    #     pass

    # block_offset = attr.ib(
    #     default=0,
    #     converter=int,
    #     metadata=graham.create_metadata(
    #         field=marshmallow.fields.Integer(),
    #     ),
    # )

    size = create_size_attribute()

    enumeration_uuid = epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
    )
    epyqlib.attrsmodel.attrib(
        attribute=enumeration_uuid,
        human_name="Enumeration",
        data_display=epyqlib.attrsmodel.name_from_uuid,
        delegate=epyqlib.attrsmodel.RootDelegateCache(
            list_selection_root="enumerations",
        ),
    )

    units = epyqlib.attrsmodel.create_str_or_none_attribute()

    # get = attr.ib(
    #     default=None,
    #     converter=epyqlib.attrsmodel.to_str_or_none,
    #     metadata=graham.create_metadata(
    #         field=marshmallow.fields.String(allow_none=True),
    #     ),
    # )
    # epyqlib.attrsmodel.attrib(
    #     attribute=get,
    #     no_column=True,
    # )
    # set = attr.ib(
    #     default=None,
    #     converter=epyqlib.attrsmodel.to_str_or_none,
    #     metadata=graham.create_metadata(
    #         field=marshmallow.fields.String(allow_none=True),
    #     ),
    # )
    # epyqlib.attrsmodel.attrib(
    #     attribute=set,
    #     no_column=True,
    # )

    # mandatory = attr.ib(
    #     default=True,
    #     converter=epyqlib.attrsmodel.two_state_checkbox,
    #     metadata=graham.create_metadata(
    #         field=marshmallow.fields.Boolean(),
    #     ),
    # )

    uuid = epyqlib.attrsmodel.attr_uuid()
    # value  doesn't make sense here, this is interface definition, not a
    #       value set

    # ? point_id = attr.ib()
    # ? index = index

    # last access time?
    # time = time

    #     getter_code = attr.ib()
    #     setter_code = attr.ib()
    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return isinstance(node, epyqlib.pm.parametermodel.Parameter)

    def child_from(self, node):
        self.parameter_uuid = node.uuid

        return None

    @epyqlib.attrsmodel.check_children
    def check(self, result, models):
        if self.parameter_uuid is None:
            result.append_child(
                epyqlib.checkresultmodel.Result(
                    node=self,
                    severity=epyqlib.checkresultmodel.ResultSeverity.error,
                    message="No parameter connected",
                )
            )
        else:
            root = self.find_root()
            parameter = root.model.node_from_uuid(self.parameter_uuid)

            # Currently no "Static Modbus..." parameters.
            # if (
            #     parameter.tree_parent.name.startswith("Static Modbus")
            #     and parameter.tree_parent.tree_parent.tree_parent is None
            # ):
            #     result.append_child(
            #         epyqlib.checkresultmodel.Result(
            #             node=self,
            #             severity=(epyqlib.checkresultmodel.ResultSeverity.information),
            #             message=(
            #                 "Connected to temporary Static Modbus imported parameter"
            #             ),
            #         )
            #     )

            if not parameter.uses_interface_item():
                result.append_child(
                    epyqlib.checkresultmodel.Result(
                        node=self,
                        severity=(epyqlib.checkresultmodel.ResultSeverity.information),
                        message=("Connected to old-style parameter"),
                    )
                )

                access_level = root.model.node_from_uuid(
                    parameter.access_level_uuid,
                )

                if access_level.value > 0:
                    result.append_child(
                        epyqlib.checkresultmodel.Result(
                            node=self,
                            severity=(epyqlib.checkresultmodel.ResultSeverity.warning),
                            message=("Access level will not be enforced"),
                        )
                    )

        return result

    can_delete = epyqlib.attrsmodel.childless_can_delete
    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    internal_move = epyqlib.attrsmodel.default_internal_move


@graham.schemify(tag="function_data_bitfield_member", register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class FunctionDataBitfieldMember(epyqlib.treenode.TreeNode):
    parameter_uuid = create_parameter_uuid_attribute()

    bit_offset = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_int_or_none,
        metadata=graham.create_metadata(
            field=marshmallow.fields.Integer(allow_none=True),
        ),
    )

    bit_length = create_size_attribute(default=1)

    type_uuid = epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
        human_name="Type",
        data_display=epyqlib.attrsmodel.name_from_uuid,
        list_selection_root="staticmodbus types",
    )

    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return isinstance(node, epyqlib.pm.parametermodel.Parameter)

    def child_from(self, node):
        self.parameter_uuid = node.uuid

        return None

    can_delete = epyqlib.attrsmodel.childless_can_delete
    all_addable_types = epyqlib.attrsmodel.empty_all_addable_types
    addable_types = epyqlib.attrsmodel.empty_addable_types
    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag="function_data_bitfield", register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class FunctionDataBitfield(epyqlib.treenode.TreeNode):
    parameter_uuid = create_parameter_uuid_attribute()

    children = attr.ib(
        factory=list,
        metadata=graham.create_metadata(
            field=graham.fields.MixedList(
                fields=(
                    marshmallow.fields.Nested(
                        graham.schema(FunctionDataBitfieldMember)
                    ),
                )
            ),
        ),
    )

    # TODO: though this only really makes sense as one of the bitfield types
    type_uuid = epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
        human_name="Type",
        data_display=epyqlib.attrsmodel.name_from_uuid,
        list_selection_root="staticmodbus types",
    )

    # block_offset = attr.ib(
    #     default=0,
    #     converter=int,
    #     metadata=graham.create_metadata(
    #         field=marshmallow.fields.Integer(),
    #     ),
    # )

    size = create_size_attribute()

    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return isinstance(
            node,
            (
                FunctionDataBitfieldMember,
                epyqlib.pm.parametermodel.Parameter,
            ),
        )

    def can_delete(self, node=None):
        if node is None:
            return self.tree_parent.can_delete(node=self)

        return True

    def child_from(self, node):
        if isinstance(node, epyqlib.pm.parametermodel.Parameter):
            self.parameter_uuid = node.uuid
            return None

        return node

    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


# Not clear if this functionality will be used in the future.
# def check_block_offsets_and_length(self):
#     length = 0
#
#     root = self.find_root()
#
#     for point in self.children:
#         try:
#             type_ = root.model.node_from_uuid(point.type_uuid)
#         except epyqlib.attrsmodel.NotFoundError:
#             point_name = root.model.node_from_uuid(point.parameter_uuid).name
#             raise TypeNotFoundError(
#                 f"Point {point_name!r}" f" has unknown type uuid {point.type_uuid!r}"
#             )
#
#         if type_.name != "string" and point.size != type_.value:
#             point_name = root.model.node_from_uuid(point.parameter_uuid).name
#             raise MismatchedSizeAndTypeError(
#                 f"Expected {type_.value} for {type_.name}"
#                 f", is {point.size} for {point_name}"
#             )
#
#         length += point.size
#
#     return length


# @graham.schemify(tag="staticmodbus_header_block", register=True)
# @epyqlib.attrsmodel.ify()
# @epyqlib.utils.qt.pyqtify()
# @attr.s(hash=False)
# class HeaderBlock(epyqlib.treenode.TreeNode):
#     name = attr.ib(
#         default="Header",
#         metadata=graham.create_metadata(
#             field=marshmallow.fields.String(),
#         ),
#     )
#     # offset = attr.ib(
#     #     default=0,
#     #     converter=int,
#     # )
#     children = attr.ib(
#         factory=list,
#         metadata=graham.create_metadata(
#             field=graham.fields.MixedList(
#                 fields=(marshmallow.fields.Nested(graham.schema(FunctionData)),)
#             ),
#         ),
#     )
#
#     uuid = epyqlib.attrsmodel.attr_uuid()
#
#     def __attrs_post_init__(self):
#         super().__init__()
#
#     def can_drop_on(self, node):
#         return False
#
#     def can_delete(self, node=None):
#         return False
#
#     check_offsets_and_length = check_block_offsets_and_length
#
#     def add_data_points(self, uint16_uuid, model_id):
#         parameters = [
#             epyqlib.pm.parametermodel.Parameter(
#                 name=model_id,
#                 abbreviation="ID",
#                 read_only=True,
#             ),
#             epyqlib.pm.parametermodel.Parameter(
#                 name="",
#                 abbreviation="L",
#                 comment="Model Length",
#                 read_only=True,
#             ),
#         ]
#         points = [
#             FunctionData(
#                 # block_offset=0,
#                 size=1,
#                 type_uuid=uint16_uuid,
#                 parameter_uuid=parameters[0].uuid,
#                 # mandatory=True,
#             ),
#             FunctionData(
#                 # block_offset=1,
#                 size=1,
#                 type_uuid=uint16_uuid,
#                 parameter_uuid=parameters[1].uuid,
#                 # mandatory=True,
#             ),
#         ]
#
#         for point in points:
#             self.append_child(point)
#
#         return parameters
#
#     remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
#     child_from = epyqlib.attrsmodel.default_child_from
#     internal_move = epyqlib.attrsmodel.default_internal_move
#     check = epyqlib.attrsmodel.check_just_children
#
#
# @graham.schemify(tag="staticmodbus_fixed_block", register=True)
# @epyqlib.attrsmodel.ify()
# @epyqlib.utils.qt.pyqtify()
# @attr.s(hash=False)
# class FixedBlock(epyqlib.treenode.TreeNode):
#     name = attr.ib(
#         default="Fixed Block",
#         metadata=graham.create_metadata(
#             field=marshmallow.fields.String(),
#         ),
#     )
#     # offset = attr.ib(
#     #     default=2,
#     #     converter=int,
#     # )
#     children = attr.ib(
#         factory=list,
#         metadata=graham.create_metadata(
#             field=graham.fields.MixedList(
#                 fields=(
#                     marshmallow.fields.Nested(graham.schema(FunctionData)),
#                     marshmallow.fields.Nested(graham.schema(FunctionDataBitfield)),
#                 )
#             ),
#         ),
#     )
#
#     uuid = epyqlib.attrsmodel.attr_uuid()
#
#     def __attrs_post_init__(self):
#         super().__init__()
#         self.pyqt_signals.child_added_complete.connect(self.update)
#         self.pyqt_signals.child_removed_complete.connect(self.update)
#
#     def can_drop_on(self, node):
#         return isinstance(
#             node,
#             (
#                 epyqlib.pm.parametermodel.Parameter,
#                 FunctionData,
#                 FunctionDataBitfield,
#             ),
#         )
#
#     def can_delete(self, node=None):
#         if node is None:
#             return self.tree_parent.can_delete(node=self)
#
#         return isinstance(node, (FunctionData, FunctionDataBitfield))
#
#     def child_from(self, node):
#         if isinstance(node, epyqlib.pm.parametermodel.Parameter):
#             return FunctionData(parameter_uuid=node.uuid)
#
#         return node
#
#     def update(self):
#         block_offset = 0
#         for pt in self.children:
#             pt.block_offset = block_offset
#             block_offset = block_offset + pt.size
#
#     check_offsets_and_length = check_block_offsets_and_length
#     remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
#     internal_move = epyqlib.attrsmodel.default_internal_move
#     check = epyqlib.attrsmodel.check_just_children


@graham.schemify(
    tag="staticmodbus_table_repeating_block_reference_function_data_reference",
    register=True,
)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@epyqlib.utils.qt.pyqtify_passthrough_properties(
    original="original",
    field_names=("parameter_uuid",),
)
@attr.s(hash=False)
class TableRepeatingBlockReferenceFunctionDataReference(epyqlib.treenode.TreeNode):
    parameter_uuid = create_parameter_uuid_attribute()

    factor_uuid = create_factor_uuid_attribute()
    original = epyqlib.attrsmodel.create_reference_attribute()

    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return False

    can_delete = epyqlib.attrsmodel.childless_can_delete
    all_addable_types = epyqlib.attrsmodel.empty_all_addable_types
    addable_types = epyqlib.attrsmodel.empty_addable_types
    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag="staticmodbus_table_repeating_block", register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@epyqlib.utils.qt.pyqtify_passthrough_properties(
    original="original",
    field_names=("name",),
)
@attr.s(hash=False)
class TableRepeatingBlockReference(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default="Table Repeating Block Reference",
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )
    # offset = attr.ib(
    #     default=2,
    #     converter=int,
    # )
    children = attr.ib(
        factory=list,
        metadata=graham.create_metadata(
            field=graham.fields.MixedList(
                fields=(
                    marshmallow.fields.Nested(
                        graham.schema(
                            TableRepeatingBlockReferenceFunctionDataReference,
                        )
                    ),
                )
            ),
        ),
    )

    original = epyqlib.attrsmodel.create_reference_attribute()

    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    @classmethod
    def all_addable_types(cls):
        return epyqlib.attrsmodel.create_addable_types(())

    @staticmethod
    def addable_types():
        return {}

    def can_drop_on(self, node):
        return False

    def can_delete(self, node=None):
        if node is None:
            return self.tree_parent.can_delete(node=self)

        return False

    # def check_offsets_and_length(self):
    #     return self.original.check_block_offsets_and_length()

    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag="staticmodbus_table_repeating_block", register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@epyqlib.utils.qt.pyqtify_passthrough_properties(
    original="original",
    field_names=("name",),
)
@attr.s(hash=False)
class FunctionDataReference(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default="Table Data Point Reference",
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )
    offset = attr.ib(
        default=2,
        converter=int,
    )
    children = attr.ib(
        factory=list,
        metadata=graham.create_metadata(
            field=graham.fields.MixedList(fields=()),
        ),
    )

    original = attr.ib(
        default=None,
        metadata=graham.create_metadata(
            field=epyqlib.attrsmodel.Reference(allow_none=True),
        ),
    )
    epyqlib.attrsmodel.attrib(
        attribute=original,
        no_column=True,
    )

    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    @classmethod
    def all_addable_types(cls):
        return epyqlib.attrsmodel.create_addable_types(())

    @staticmethod
    def addable_types():
        return {}

    def can_drop_on(self, node):
        return False

    def can_delete(self, node=None):
        return False

    # check_offsets_and_length = check_block_offsets_and_length
    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from


@graham.schemify(tag="table_model_reference", register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class TableRepeatingBlock(epyqlib.treenode.TreeNode):
    uuid = epyqlib.attrsmodel.attr_uuid()
    name = attr.ib(
        default="Table Reference",
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )

    # abbreviation = epyqlib.pm.parametermodel.create_abbreviation_attribute()

    children = attr.ib(
        factory=list,
        metadata=graham.create_metadata(
            field=graham.fields.MixedList(
                fields=(marshmallow.fields.Nested(graham.schema(FunctionData)),)
            ),
        ),
    )

    # offset = attr.ib(
    #     default=2,
    #     converter=int,
    # )

    repeats = attr.ib(
        default=0,
        converter=int,
    )

    path = attr.ib(
        factory=tuple,
    )
    epyqlib.attrsmodel.attrib(
        attribute=path,
        no_column=True,
    )
    graham.attrib(
        attribute=path,
        field=graham.fields.Tuple(marshmallow.fields.UUID()),
    )

    def __attrs_post_init__(self):
        super().__init__()

    @classmethod
    def all_addable_types(cls):
        return epyqlib.attrsmodel.create_addable_types(())

    @staticmethod
    def addable_types():
        return {}

    def can_drop_on(self, node):
        return False

    def can_delete(self, node=None):
        if node is None:
            return self.tree_parent.can_delete(node=self)

        return False

    # def check_block_offsets_and_length(self):
    #     return self.repeats * check_block_offsets_and_length(self)

    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag="table", register=True)
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class Table(epyqlib.treenode.TreeNode):
    name = attr.ib(
        default="New Table",
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )

    parameter_table_uuid = epyqlib.attrsmodel.attr_uuid(
        default=None,
        allow_none=True,
    )
    epyqlib.attrsmodel.attrib(
        attribute=parameter_table_uuid,
        human_name="Table UUID",
    )

    children = attr.ib(
        default=attr.Factory(list),
        metadata=graham.create_metadata(
            field=graham.fields.MixedList(
                fields=(
                    marshmallow.fields.Nested(graham.schema(TableRepeatingBlock)),
                    marshmallow.fields.Nested(graham.schema(FunctionData)),
                )
            ),
        ),
    )

    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    @classmethod
    def all_addable_types(cls):
        return epyqlib.attrsmodel.create_addable_types(())

    @staticmethod
    def addable_types():
        return {}

    @staticmethod
    def can_drop_on(node):
        return isinstance(node, epyqlib.pm.parametermodel.Table)

    def can_delete(self, node=None):
        if node is None:
            return self.tree_parent.can_delete(node=self)

        return True

    def child_from(self, node):
        self.parameter_table_uuid = node.uuid
        return None

    def update(self, table=None):
        old_nodes = self.recursively_remove_children()
        old_nodes_by_path = {
            getattr(node, "path", getattr(node, "parameter_uuid", node.uuid)): node
            for node in old_nodes
        }

        if self.parameter_table_uuid is None:
            return

        root = self.find_root()
        model = root.model

        if table is None:
            table = model.node_from_uuid(self.parameter_table_uuid)
        elif table.uuid != self.table_uuid:
            raise ConsistencyError()

        master_array_function_data_by_uuid = {}

        for section in table.arrays_and_groups:
            if isinstance(section, epyqlib.pm.parametermodel.Array):
                array_element = section.children[0]
                node = old_nodes_by_path.get(array_element.uuid)
                if node is None:
                    node = FunctionData(
                        parameter_uuid=array_element.uuid,
                    )
                self.append_child(node)
                master_array_function_data_by_uuid[array_element.uuid] = node
            elif isinstance(section, epyqlib.pm.parametermodel.Group):
                for element in section.children:
                    node = old_nodes_by_path.get(element.uuid)
                    if node is None:
                        node = FunctionData(
                            parameter_uuid=element.uuid,
                        )
                    self.append_child(node)
                    master_array_function_data_by_uuid[element.uuid] = node

        for combination in table.combinations:
            not_first_curve = any(
                (
                    layer.tree_parent.name == "Curves"
                    and layer.name != layer.tree_parent.children[0].name
                )
                for layer in combination
            )
            if not_first_curve:
                continue

            curve_enumeration_search = [
                layer for layer in combination if layer.tree_parent.name == "Curves"
            ]
            if len(curve_enumeration_search) == 0:
                curve_count = 0
            elif len(curve_enumeration_search) == 1:
                curve_enumeration = curve_enumeration_search[0].tree_parent
                curve_count = len(curve_enumeration.children)
            else:
                raise blue

            base_path = tuple(node.uuid for node in combination)

            block_node = old_nodes_by_path.get(base_path)

            if block_node is None:
                block_node = TableRepeatingBlock(
                    name=" - ".join(
                        item.name
                        for item in combination
                        if item.tree_parent.name != "Curves"
                    ),
                    path=base_path,
                )

            block_node.repeats = curve_count

            self.append_child(block_node)

            (in_tree,) = table.group.nodes_by_attribute(
                attribute_value=tuple(node.uuid for node in combination),
                attribute_name="path",
            )
            # continue

            # block_offset = 0

            group_elements = [[], []]
            group_of_groups = group_elements[0]

            for child in in_tree.children:
                if isinstance(child.original, epyqlib.pm.parametermodel.Array):
                    group_of_groups = group_elements[1]
                    continue

                group_of_groups.append(child.children)

            for group in group_elements:
                group[:] = itertools.chain.from_iterable(group)

            # TODO: CAMPid 143707880547014313476753071297360068134
            for element in group_elements[0]:
                point_node = old_nodes_by_path.get(element.uuid)
                reference_function_data = master_array_function_data_by_uuid[
                    element.original.uuid
                ]
                if point_node is None:
                    point_node = FunctionData(
                        parameter_uuid=element.uuid,
                    )
                # point_node.mandatory = reference_function_data.mandatory
                point_node.units = reference_function_data.units
                point_node.type_uuid = reference_function_data.type_uuid
                point_node.size = reference_function_data.size
                point_node.enumeration_uuid = reference_function_data.enumeration_uuid
                # point_node.block_offset = block_offset
                block_node.append_child(point_node)
                # block_offset += point_node.size

            array_elements = itertools.chain.from_iterable(
                zip(
                    *(
                        array.children
                        for array in in_tree.children
                        if isinstance(array.original, epyqlib.pm.parametermodel.Array)
                    )
                ),
            )
            for element in array_elements:
                point_node = old_nodes_by_path.get(element.uuid)
                reference_function_data = master_array_function_data_by_uuid[
                    element.tree_parent.children[0].original.uuid
                ]
                if point_node is None:
                    point_node = FunctionData(
                        parameter_uuid=element.uuid,
                    )
                # point_node.mandatory = reference_function_data.mandatory
                point_node.units = reference_function_data.units
                point_node.type_uuid = reference_function_data.type_uuid
                point_node.size = reference_function_data.size
                point_node.enumeration_uuid = reference_function_data.enumeration_uuid
                # point_node.block_offset = block_offset
                block_node.append_child(point_node)
                # block_offset += point_node.size

            # TODO: CAMPid 143707880547014313476753071297360068134
            for element in group_elements[1]:
                point_node = old_nodes_by_path.get(element.uuid)
                reference_function_data = master_array_function_data_by_uuid[
                    element.original.uuid
                ]
                if point_node is None:
                    point_node = FunctionData(
                        parameter_uuid=element.uuid,
                    )
                # point_node.mandatory = reference_function_data.mandatory
                point_node.units = reference_function_data.units
                point_node.type_uuid = reference_function_data.type_uuid
                point_node.size = reference_function_data.size
                point_node.enumeration_uuid = reference_function_data.enumeration_uuid
                # point_node.block_offset = block_offset
                block_node.append_child(point_node)
                # block_offset += point_node.size

    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


# @graham.schemify(tag="staticmodbus_model", register=True)
# @epyqlib.attrsmodel.ify()
# @epyqlib.utils.qt.pyqtify()
# @attr.s(hash=False)
# class Model(epyqlib.treenode.TreeNode):
#     name = attr.ib(
#         default=0,
#         converter=str,
#         metadata=graham.create_metadata(
#             field=marshmallow.fields.String(),
#         ),
#     )
#     # id = attr.ib(
#     #     default=0,
#     #     converter=int,
#     #     metadata=graham.create_metadata(
#     #         field=marshmallow.fields.Integer(),
#     #     ),
#     # )  # 103
#     length = attr.ib(
#         default=0,
#         converter=int,
#         metadata=graham.create_metadata(
#             field=marshmallow.fields.Integer(),
#         ),
#     )  # 50, in the spirit of over constraining like how
#     # data points have their offset
#
#     # ?
#     # self.model_id = '{}'.format(model_id)
#     # self.namespace = namespace
#     # self.index = index
#
#     # special children
#     # header: id and length
#
#     # self.point_data = []
#     children = attr.ib(
#         factory=lambda: [HeaderBlock(), FixedBlock()],
#         metadata=graham.create_metadata(
#             field=graham.fields.MixedList(
#                 fields=(
#                     # marshmallow.fields.Nested(graham.schema(HeaderBlock)),
#                     # marshmallow.fields.Nested(graham.schema(FixedBlock)),
#                     marshmallow.fields.Nested(graham.schema(FunctionData)),
#                     marshmallow.fields.Nested(
#                         graham.schema(TableRepeatingBlockReference)
#                     ),
#                 )
#             ),
#         ),
#     )
#
#     uuid = epyqlib.attrsmodel.attr_uuid()
#
#     def __attrs_post_init__(self):
#         super().__init__()
#
#     def remove_old_on_drop(self, node):
#         return False
#
#     def child_from(self, node):
#         reference = TableRepeatingBlockReference(original=node)
#         for child in node.tree_parent.children:
#             if not isinstance(child, FunctionData):
#                 continue
#
#             reference.append_child(
#                 TableRepeatingBlockReferenceFunctionDataReference(
#                     original=child,
#                 ),
#             )
#
#         return reference
#
#     def can_drop_on(self, node):
#         return isinstance(node, TableRepeatingBlock) and not any(
#             isinstance(child, TableRepeatingBlockReference) for child in self.children
#         )
#
#     def can_delete(self, node=None):
#         if node is None:
#             return self.tree_parent.can_delete(node=self)
#
#         return not isinstance(node, (HeaderBlock, FixedBlock))
#
#     def check_offsets_and_length(self):
#         length = 0
#
#         for block in self.children:
#             length += block.check_offsets_and_length()
#
#         return length
#
#     internal_move = epyqlib.attrsmodel.default_internal_move
#     check = epyqlib.attrsmodel.check_just_children


# class Repeating...?


Root = epyqlib.attrsmodel.Root(
    default_name="Static Modbus",
    valid_types=(
        FunctionData,
        Table,
        TableRepeatingBlockReferenceFunctionDataReference,
        FunctionDataBitfield,
        FunctionDataBitfieldMember,
    ),
    # valid_types=(Model, Table),
)


types = epyqlib.attrsmodel.Types(
    types=(
        Root,
        # Model,
        FunctionData,
        FunctionDataBitfield,
        FunctionDataBitfieldMember,
        # HeaderBlock,
        # FixedBlock,
        Table,
        TableRepeatingBlock,
        TableRepeatingBlockReference,
        TableRepeatingBlockReferenceFunctionDataReference,
    ),
)


# TODO: CAMPid 943896754217967154269254167
def merge(name, *types):
    return tuple((x, name) for x in types)


columns = epyqlib.attrsmodel.columns(
    (
        merge(
            "name",
            # HeaderBlock,
            # FixedBlock,
            Table,
            TableRepeatingBlock,
            TableRepeatingBlockReference,
        )
        # + merge("name", Model)
        # + merge("id", Model)
        + merge(
            "parameter_uuid",
            FunctionData,
            TableRepeatingBlockReferenceFunctionDataReference,
            FunctionDataBitfield,
            FunctionDataBitfieldMember,
        )
    ),
    # merge("abbreviation", TableRepeatingBlock),
    merge("not_implemented", FunctionData),
    merge("size", FunctionData, FunctionDataBitfield),
    # merge("length", Model) + merge("size", FunctionData, FunctionDataBitfield),
    merge("repeats", TableRepeatingBlock),
    # merge("hand_coded_getter", FunctionData),
    # merge("hand_coded_setter", FunctionData),
    merge(
        "factor_uuid",
        FunctionData,
        TableRepeatingBlockReferenceFunctionDataReference,
    ),
    merge("units", FunctionData),
    merge("enumeration_uuid", FunctionData),
    merge("type_uuid", FunctionData, FunctionDataBitfield, FunctionDataBitfieldMember),
    merge("bit_length", FunctionDataBitfieldMember),
    merge("bit_offset", FunctionDataBitfieldMember),
    merge("parameter_table_uuid", Table),
    # merge("mandatory", FunctionData),
    # merge(
    #     "offset",
    #     FunctionData,
    #     HeaderBlock,
    #     FixedBlock,
    #     TableRepeatingBlockReference,
    #     TableRepeatingBlock,
    # ),
    # merge("block_offset", FunctionData, FunctionDataBitfield),
    merge("uuid", *types.types.values()),
)


# TODO: CAMPid 075454679961754906124539691347967
@attr.s
class ReferencedUuidNotifier:
    changed = epyqlib.utils.qt.Signal("PyQt_PyObject")

    view = attr.ib(default=None)
    selection_model = attr.ib(default=None)

    def __attrs_post_init__(self):
        if self.view is not None:
            self.set_view(self.view)

    def set_view(self, view):
        self.disconnect_view()

        self.view = view
        self.selection_model = self.view.selectionModel()
        self.selection_model.currentChanged.connect(
            self.current_changed,
        )

    def disconnect_view(self):
        if self.selection_model is not None:
            self.selection_model.currentChanged.disconnect(
                self.current_changed,
            )
        self.view = None
        self.selection_model = None

    def current_changed(self, current, previous):
        if not current.isValid():
            return

        index = epyqlib.utils.qt.resolve_index_to_model(
            index=current,
        )
        model = index.data(epyqlib.utils.qt.UserRoles.attrs_model)
        node = model.node_from_index(index)

        parameter_uuid = getattr(node, "parameter_uuid", None)

        if parameter_uuid is not None:
            self.changed.emit(parameter_uuid)



import itertools
import math

import attr
import openpyxl

import epyqlib.pm.parametermodel
import epyqlib.utils.general

import epcpm.c
import epcpm.sunspecmodel


builders = epyqlib.utils.general.TypeMap()
enumeration_builders = epyqlib.utils.general.TypeMap()
enumerator_builders = epyqlib.utils.general.TypeMap()


data_point_fields = attr.fields(epcpm.sunspecmodel.DataPoint)
epc_enumerator_fields = attr.fields(
    epyqlib.pm.parametermodel.SunSpecEnumerator,
)
bitfield_fields = attr.fields(epcpm.sunspecmodel.DataPointBitfield)


def attr_fill(cls, value):
    return cls(**{field.name: value for field in attr.fields(cls) if field.init})


@attr.s
class Fields:
    field_type = attr.ib(default=None)
    applicable_point = attr.ib(default=None)
    address_offset = attr.ib(default=None)
    block_offset = attr.ib(default=None)
    size = attr.ib(default=None)
    name = attr.ib(default=None)
    label = attr.ib(default=None)
    value = attr.ib(default=None)
    type = attr.ib(default=None)
    units = attr.ib(default=None)
    scale_factor = attr.ib(default=None)
    read_write = attr.ib(default=None)
    mandatory = attr.ib(default=None)
    description = attr.ib(default=None)
    notes = attr.ib(default=None)
    modbus_address = attr.ib(default=None)
    get = attr.ib(default=None)
    set = attr.ib(default=None)
    item = attr.ib(default=None)

    def as_filtered_tuple(self, filter_):
        return tuple(
            value for value, f in zip(attr.astuple(self), attr.astuple(filter_)) if f
        )


field_names = Fields(
    field_type="Field Type",
    applicable_point="Applicable Point",
    address_offset="Address Offset",
    block_offset="Block Offset",
    size="Size",
    name="Name",
    label="Label",
    value="Value",
    type="Type",
    units="Units",
    scale_factor="SF",
    read_write="R/W",
    mandatory="Mandatory M/O",
    description="Description",
    notes="Notes",
    modbus_address="Modbus Address",
    get="get",
    set="set",
    item="item",
)


point_fields = Fields(
    block_offset=data_point_fields.block_offset,
    size=data_point_fields.size,
    type=data_point_fields.type_uuid,
    scale_factor=data_point_fields.factor_uuid,
    units=data_point_fields.units,
)

bitfield_fields = Fields(
    block_offset=bitfield_fields.block_offset,
    size=bitfield_fields.size,
    type=bitfield_fields.type_uuid,
)

enumerator_fields = Fields(
    name=epc_enumerator_fields.abbreviation,
    label=epc_enumerator_fields.label,
    description=epc_enumerator_fields.description,
    notes=epc_enumerator_fields.notes,
    value=epc_enumerator_fields.value,
    field_type=epc_enumerator_fields.type,
)


def export(
    path, sunspec_model, parameters_model, column_filter=None, skip_sunspec=False
):
    if column_filter is None:
        column_filter = attr_fill(Fields, True)

    builder = epcpm.sunspectoxlsx.builders.wrap(
        wrapped=sunspec_model.root,
        parameter_uuid_finder=sunspec_model.node_from_uuid,
        parameter_model=parameters_model,
        skip_sunspec=skip_sunspec,
        column_filter=column_filter,
    )

    workbook = builder.gen()

    path.parent.mkdir(parents=True, exist_ok=True)
    workbook.save(path)


@builders(epcpm.sunspecmodel.Root)
@attr.s
class Root:
    wrapped = attr.ib()
    column_filter = attr.ib()
    skip_sunspec = attr.ib(default=False)
    parameter_uuid_finder = attr.ib(default=None)
    parameter_model = attr.ib(default=None)
    sort_models = attr.ib(default=False)

    def gen(self):
        workbook = openpyxl.Workbook()
        workbook.remove(workbook.active)

        workbook.create_sheet("License Agreement")
        workbook.create_sheet("Summary")
        workbook.create_sheet("Index")

        if not self.skip_sunspec:
            if self.sort_models:
                children = sorted(
                    self.wrapped.children,
                    key=lambda child: (
                        0 if isinstance(child, epcpm.sunspecmodel.Model) else 1,
                        getattr(child, "id", math.inf),
                    ),
                )
            else:
                children = list(self.wrapped.children)

            model_offset = 2  # account for starting 'SunS'
            for model in children:
                if isinstance(model, epcpm.sunspecmodel.Table):
                    # TODO: for now, implement it soon...
                    continue

                worksheet = workbook.create_sheet()

                model_offset += builders.wrap(
                    wrapped=model,
                    worksheet=worksheet,
                    padding_type=self.parameter_model.list_selection_roots[
                        "sunspec types"
                    ].child_by_name("pad"),
                    parameter_uuid_finder=self.parameter_uuid_finder,
                    column_filter=self.column_filter,
                    model_offset=model_offset,
                ).gen()

        return workbook


@builders(epcpm.sunspecmodel.Model)
@attr.s
class Model:
    wrapped = attr.ib()
    worksheet = attr.ib()
    padding_type = attr.ib()
    column_filter = attr.ib()
    model_offset = attr.ib()  # starting Modbus address for the model
    parameter_uuid_finder = attr.ib(default=None)

    def gen(self):
        self.worksheet.title = str(self.wrapped.id)
        self.worksheet.append(field_names.as_filtered_tuple(self.column_filter))

        self.wrapped.children[0].check_offsets_and_length()

        overall_length = sum(
            child.check_offsets_and_length() for child in self.wrapped.children[1:]
        )

        add_padding = (overall_length % 2) == 1
        if add_padding:
            overall_length += 1

        accumulated_length = 0

        rows = []

        model_types = ["Header", "Fixed Block", "Repeating Block"]

        zipped = zip(enumerate(self.wrapped.children), model_types)
        for (i, child), model_type in zipped:
            builder = builders.wrap(
                wrapped=child,
                add_padding=add_padding and i == 1,
                padding_type=self.padding_type,
                model_type=model_type,
                model_id=self.wrapped.id,
                model_offset=self.model_offset,
                parameter_uuid_finder=self.parameter_uuid_finder,
                address_offset=accumulated_length,
            )

            built_rows, block_length = builder.gen()
            accumulated_length += block_length
            if len(built_rows) > 0:
                rows.extend(built_rows)
                rows.append(Fields())

        for i, row in enumerate(rows):
            if i == 0:
                row.value = self.wrapped.id
            elif i == 1:
                row.value = overall_length

            self.worksheet.append(row.as_filtered_tuple(self.column_filter))

        for block in self.wrapped.children:
            builder = enumeration_builders.wrap(
                wrapped=block,
                parameter_uuid_finder=self.parameter_uuid_finder,
            )
            rows = builder.gen()

            for row in rows:
                self.worksheet.append(
                    row.as_filtered_tuple(self.column_filter),
                )

        return overall_length + 2  # add header length


@builders(epcpm.sunspecmodel.Table)
@attr.s
class Table:
    wrapped = attr.ib()
    worksheet = attr.ib()
    padding_type = attr.ib()
    column_filter = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    def gen(self):
        return []


@enumeration_builders(epcpm.sunspecmodel.DataPoint)
@attr.s
class Enumeration:
    wrapped = attr.ib()
    point = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    def gen(self):
        rows = []

        enumeration_uuid = getattr(self.wrapped, "enumeration_uuid", None)
        if enumeration_uuid is None:
            return rows

        enumeration = self.parameter_uuid_finder(enumeration_uuid)

        enumerators_by_bit = {
            enumerator.value: enumerator for enumerator in enumeration.children
        }

        # 16 bits per register
        total_bit_count = self.wrapped.size * 16
        decimal_digits = len(str(total_bit_count - 1))

        for bit in range(total_bit_count):
            enumerator = enumerators_by_bit.get(bit)
            if enumerator is None:
                padded_bit_string = f"{bit:0{decimal_digits}}"
                enumerator = epyqlib.pm.parametermodel.SunSpecEnumerator(
                    label=f"Reserved - {padded_bit_string}",
                    name=f"Rsvd{padded_bit_string}",
                    value=bit,
                )

            builder = enumerator_builders.wrap(
                wrapped=enumerator,
                point=self.point,
                parameter_uuid_finder=self.parameter_uuid_finder,
            )
            rows.append(builder.gen())

        return rows


@enumerator_builders(epyqlib.pm.parametermodel.SunSpecEnumerator)
@attr.s
class Enumerator:
    wrapped = attr.ib()
    point = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    # TODO: CAMPid 07397546759269756456100183066795496952476951653
    def gen(self):
        row = Fields()

        for name, field in attr.asdict(enumerator_fields).items():
            if field is None:
                continue

            setattr(row, name, getattr(self.wrapped, field.name))

        if row.name is None:
            row.name = self.wrapped.name

        parameter = self.parameter_uuid_finder(self.point.parameter_uuid)
        row.applicable_point = parameter.abbreviation

        field_type_parameter = self.parameter_uuid_finder(self.point.type_uuid)
        row.field_type = field_type_parameter.name

        return row


def build_uuid_scale_factor_dict(points, parameter_uuid_finder):
    scale_factor_from_uuid = {}
    for point in points:
        if point.type_uuid is None:
            continue

        type_node = parameter_uuid_finder(point.type_uuid)

        if type_node is None:
            continue

        if type_node.name == "sunssf":
            scale_factor_from_uuid[point.uuid] = point

    return scale_factor_from_uuid


@builders(epcpm.sunspecmodel.TableRepeatingBlock)
@builders(epcpm.sunspecmodel.HeaderBlock)
@builders(epcpm.sunspecmodel.FixedBlock)
@attr.s
class Block:
    wrapped = attr.ib()
    add_padding = attr.ib()
    padding_type = attr.ib()
    model_type = attr.ib()
    model_id = attr.ib()
    model_offset = attr.ib()
    address_offset = attr.ib()
    repeating_block_reference = attr.ib(default=None)
    parameter_uuid_finder = attr.ib(default=None)
    is_table = attr.ib(default=False)

    def gen(self):
        # TODO: CAMPid 07548795421667967542697543743987

        scale_factor_from_uuid = build_uuid_scale_factor_dict(
            points=self.wrapped.children,
            parameter_uuid_finder=self.parameter_uuid_finder,
        )

        rows = []

        points = list(self.wrapped.children)

        if self.add_padding:
            point = epcpm.sunspecmodel.DataPoint(
                type_uuid=self.padding_type.uuid,
                block_offset=(
                    self.wrapped.children[-1].block_offset
                    + self.wrapped.children[-1].size
                ),
                size=self.padding_type.value,
            )
            # TODO: ack!  just to get the address offset calculated but
            #       not calling append_child() because i don't want to shove
            #       this into the model.  :[
            point.tree_parent = self.wrapped
            points.append(point)

        summed_increments = 0

        for child in points:
            builder = builders.wrap(
                wrapped=child,
                model_type=self.model_type,
                scale_factor_from_uuid=scale_factor_from_uuid,
                parameter_uuid_finder=self.parameter_uuid_finder,
                model_id=self.model_id,
                model_offset=self.model_offset,
                is_table=self.is_table,
                repeating_block_reference=self.repeating_block_reference,
                address_offset=self.address_offset + summed_increments,
            )
            built_rows, address_offset_increment = builder.gen()
            summed_increments += address_offset_increment
            rows.append(built_rows)

        return rows, summed_increments


@builders(epcpm.sunspecmodel.DataPointBitfield)
@attr.s
class DataPointBitfield:
    wrapped = attr.ib()
    model_type = attr.ib()
    scale_factor_from_uuid = attr.ib()
    parameter_uuid_finder = attr.ib()
    model_id = attr.ib()
    model_offset = attr.ib()
    is_table = attr.ib()
    repeating_block_reference = attr.ib()
    address_offset = attr.ib()

    def gen(self):
        row = Fields()
        row.address_offset = self.address_offset
        row.modbus_address = self.model_offset + self.address_offset
        row.field_type = self.model_type

        for name, field in attr.asdict(bitfield_fields).items():
            if field is None:
                continue

            setattr(row, name, getattr(self.wrapped, field.name))

        if row.type is not None:
            row.type = self.parameter_uuid_finder(row.type).name

        if self.wrapped.parameter_uuid is not None:
            parameter = self.parameter_uuid_finder(self.wrapped.parameter_uuid)

            row.label = parameter.name
            row.name = parameter.abbreviation
            row.notes = "" if parameter.notes is None else parameter.notes
            row.notes = f"{row.notes}  <uuid:{parameter.uuid}>".strip()

            if row.units is None:
                row.units = parameter.units
            row.description = parameter.comment
            row.read_write = "R" if parameter.read_only else "RW"

        uses_interface_item = (
            isinstance(parameter, epyqlib.pm.parametermodel.Parameter)
            and parameter.uses_interface_item()
        )

        getter = []
        setter = []

        # TODO: should we just require that it does and assume etc?
        if uses_interface_item:
            # TODO: CAMPid 9685439641536675431653179671436
            parameter_uuid = str(parameter.uuid).replace("-", "_")
            item_name = f"interfaceItem_{parameter_uuid}"

            getter.extend(
                [
                    f"{item_name}.common.sunspec.getter(",
                    [
                        f"(InterfaceItem_void *) &{item_name},",
                        f"Meta_Value",
                    ],
                    f");",
                ]
            )
            setter.extend(
                [
                    f"{item_name}.common.sunspec.setter(",
                    [
                        f"(InterfaceItem_void *) &{item_name},",
                        f"true,",
                        f"Meta_Value",
                    ],
                    f");",
                ]
            )

        if len(getter) > 0:
            # TODO: what if write-only?
            row.get = epcpm.c.format_nested_lists(getter)

        if not parameter.read_only:
            row.set = epcpm.c.format_nested_lists(setter)
        else:
            row.set = None

        return row, row.size


@enumeration_builders(epcpm.sunspecmodel.DataPointBitfield)
@attr.s
class DataPointBitfield:
    wrapped = attr.ib()
    point = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    def gen(self):
        rows = []

        # 16 bits per register
        total_bit_count = self.wrapped.size * 16
        decimal_digits = len(str(total_bit_count - 1))

        enumerators_by_bit = {
            enumerator.bit_offset
            + i: [enumerator, i if enumerator.bit_length > 1 else None]
            for enumerator in self.wrapped.children
            for i in range(enumerator.bit_length)
        }

        member_parameter = self.parameter_uuid_finder(
            self.wrapped.parameter_uuid,
        )

        for bit in range(total_bit_count):
            enumerator, index = enumerators_by_bit.get(bit, [None, None])

            if enumerator is None:
                padded_bit_string = f"{bit:0{decimal_digits}}"
                row = Fields(
                    field_type=f"bitfield{total_bit_count}",
                    value=bit,
                    applicable_point=member_parameter.abbreviation,
                    name=f"Rsvd{padded_bit_string}",
                    label=f"Reserved - {padded_bit_string}",
                )
            else:
                builder = enumerator_builders.wrap(
                    wrapped=enumerator,
                    point=self.point,
                    parameter_uuid_finder=self.parameter_uuid_finder,
                )
                row = builder.gen()

                if index is not None:
                    padded_index_string = f"{index:0{decimal_digits}}"

                    row = attr.evolve(
                        row,
                        name=f"{row.name}{padded_index_string}",
                        label=f"{row.label} - {padded_index_string}",
                        value=row.value + index,
                    )

            rows.append(row)

        return rows


@enumerator_builders(epcpm.sunspecmodel.DataPointBitfieldMember)
@attr.s
class DataPointBitfieldMember:
    wrapped = attr.ib()
    point = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    # TODO: CAMPid 07397546759269756456100183066795496952476951653
    def gen(self):
        field_parameter = self.parameter_uuid_finder(self.point.parameter_uuid)
        member_parameter = self.parameter_uuid_finder(self.wrapped.parameter_uuid)

        length = self.point.size * 16

        row = Fields(
            field_type=f"bitfield{length}",
            value=self.wrapped.bit_offset,
            applicable_point=field_parameter.abbreviation,
            name=member_parameter.abbreviation,
            label=member_parameter.name,
            description=member_parameter.comment,
            notes=member_parameter.notes,
        )

        return row


@enumeration_builders(epcpm.sunspecmodel.HeaderBlock)
@enumeration_builders(epcpm.sunspecmodel.FixedBlock)
@enumeration_builders(epcpm.sunspecmodel.TableRepeatingBlockReference)
@attr.s
class GenericEnumeratorBuilder:
    wrapped = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    def gen(self):
        rows = []

        for child in self.wrapped.children:
            builder = enumeration_builders.wrap(
                wrapped=child,
                point=child,
                parameter_uuid_finder=self.parameter_uuid_finder,
            )

            new_rows = builder.gen()

            if len(new_rows) > 0:
                rows.extend(new_rows)
                rows.append(Fields())

        return rows


@enumeration_builders(
    epcpm.sunspecmodel.TableRepeatingBlockReferenceDataPointReference,
)
@attr.s
class TableRepeatingBlockReferenceDataPointReferenceEnumerationBuilder:
    wrapped = attr.ib()
    point = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    def gen(self):
        builder = enumeration_builders.wrap(
            wrapped=self.wrapped.original,
            point=self.point.original,
            parameter_uuid_finder=self.parameter_uuid_finder,
        )

        return builder.gen()


@builders(epcpm.sunspecmodel.TableRepeatingBlockReference)
@attr.s
class TableRepeatingBlockReference:
    wrapped = attr.ib()
    add_padding = attr.ib()
    padding_type = attr.ib()
    model_type = attr.ib()
    model_id = attr.ib()
    model_offset = attr.ib()
    address_offset = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    def gen(self):
        builder = builders.wrap(
            wrapped=self.wrapped.original,
            model_type=self.model_type,
            parameter_uuid_finder=self.parameter_uuid_finder,
            add_padding=self.add_padding,
            padding_type=self.padding_type,
            model_id=self.model_id,
            model_offset=self.model_offset,
            is_table=True,
            repeating_block_reference=self.wrapped,
            address_offset=self.address_offset,
        )

        return builder.gen()


@builders(epcpm.sunspecmodel.DataPoint)
@attr.s
class Point:
    wrapped = attr.ib()
    scale_factor_from_uuid = attr.ib()
    model_type = attr.ib()
    model_id = attr.ib()
    model_offset = attr.ib()
    is_table = attr.ib()
    address_offset = attr.ib()
    repeating_block_reference = attr.ib()
    parameter_uuid_finder = attr.ib(default=None)

    # TODO: CAMPid 07397546759269756456100183066795496952476951653
    def gen(self):
        row = Fields()
        row.address_offset = self.address_offset
        row.modbus_address = self.model_offset + self.address_offset

        for name, field in attr.asdict(point_fields).items():
            if field is None:
                continue

            setattr(row, name, getattr(self.wrapped, field.name))

        if self.repeating_block_reference is not None:
            target = self.parameter_uuid_finder(self.wrapped.parameter_uuid).original
            is_array_element = isinstance(
                target,
                epyqlib.pm.parametermodel.ArrayParameterElement,
            )
            if is_array_element:
                target = target.tree_parent.children[0]

            references = [
                child
                for child in self.repeating_block_reference.children
                if target.uuid == child.parameter_uuid
            ]
            if len(references) > 0:
                (reference,) = references

                if reference.factor_uuid is not None:
                    row.scale_factor = self.parameter_uuid_finder(
                        self.parameter_uuid_finder(reference.factor_uuid).parameter_uuid
                    ).abbreviation
        else:
            if row.scale_factor is not None:
                row.scale_factor = self.parameter_uuid_finder(
                    self.scale_factor_from_uuid[row.scale_factor].parameter_uuid
                ).abbreviation

        if row.type is not None:
            row.type = self.parameter_uuid_finder(row.type).name

        row.mandatory = "M" if self.wrapped.mandatory else "O"

        if self.wrapped.parameter_uuid is not None:
            parameter = self.parameter_uuid_finder(self.wrapped.parameter_uuid)

            row.label = parameter.name
            row.name = parameter.abbreviation
            row.notes = "" if parameter.notes is None else parameter.notes
            to_tree = list(
                itertools.takewhile(
                    lambda node: not isinstance(node, epyqlib.pm.parametermodel.Table),
                    parameter.ancestors(),
                )
            )
            tree = to_tree[-1]
            if isinstance(tree, epyqlib.pm.parametermodel.TableGroupElement):
                # naturally sorted by traversal...  hopefully
                all_of_them = tree.nodes_by_filter(
                    filter=lambda node: node.original == parameter.original,
                    collection=[],
                )
                uuids_string = " ".join(
                    (f"<uuid:{parameter.uuid}>" for parameter in all_of_them),
                )
                row.notes = f"{row.notes}  {uuids_string}".strip()
            else:
                row.notes = f"{row.notes}  <uuid:{parameter.uuid}>".strip()

            if row.units is None:
                row.units = parameter.units
            row.description = parameter.comment
            row.read_write = "R" if parameter.read_only else "RW"

            meta = "[Meta_Value]"

            getter = []
            setter = []

            uses_interface_item = (
                isinstance(parameter, epyqlib.pm.parametermodel.Parameter)
                and parameter.uses_interface_item()
            )

            hand_coded_getter_function_name = getter_name(
                parameter=parameter,
                model_id=self.model_id,
                is_table=self.is_table,
            )

            hand_coded_setter_function_name = setter_name(
                parameter=parameter,
                model_id=self.model_id,
                is_table=self.is_table,
            )

            if not uses_interface_item and not self.wrapped.not_implemented:
                if row.scale_factor is not None:
                    scale_factor_updater_name = (
                        f"getSUNSPEC_MODEL{self.model_id}_{row.scale_factor}"
                    )

                    f = f"{scale_factor_updater_name}();"
                    get_scale_factor = f.format(
                        model_id=self.model_id,
                        abbreviation=row.scale_factor,
                    )
                    getter.append(get_scale_factor)
                    setter.append(get_scale_factor)

                getter.append(f"{hand_coded_getter_function_name}();")

            sunspec_model_variable = f"sunspecInterface.model{self.model_id}"

            sunspec_variable = f"{sunspec_model_variable}.{parameter.abbreviation}"

            if row.type == "pad":
                getter.append(f"{sunspec_variable} = 0x8000;")
            elif self.wrapped.not_implemented:
                value = {
                    "int16": "INT16_C(0x8000)",
                    "uint16": "UINT16_C(0xffff)",
                    "acc16": "UINT16_C(0x0000)",
                    "enum16": "UINT16_C(0xffff)",
                    "bitfield16": "UINT16_C(0xffff)",
                    "int32": "sunspecInt32ToSS32_returns(INT32_C(0x80000000))",
                    "uint32": "sunspecUint32ToSSU32_returns(UINT32_C(0xffffffff))",
                    "acc32": "sunspecUint32ToSSU32_returns(UINT32_C(0x00000000))",
                    "enum32": "sunspecUint32ToSSU32_returns(UINT32_C(0xffffffff))",
                    "bitfield32": "sunspecUint32ToSSU32_returns(UINT32_C(0xffffffff))",
                    "ipaddr": "sunspecUint32ToSSU32_returns(UINT32_C(0x00000000))",
                    "int64": "sunspecInt64ToSS64_returns(INT64_C(0x8000000000000000))",
                    # yes, acc64 seems to be an int64, not a uint64
                    "acc64": "sunspecInt64ToSS64_returns(INT64_C(0x0000000000000000))",
                    # 'ipv6addr': 'INT128_C(0x00000000000000000000000000000000)',
                    # 'float32': 'NAN',
                    "sunssf": "INT16_C(0x8000)",
                    "string": "UINT16_C(0x0000)",
                }[row.type]
                if row.type == "string":
                    getter.extend(
                        [
                            f"for (size_t i = 0; i < LENGTHOF({sunspec_variable}); i++) {{",
                            [f"{sunspec_variable}[i] = {value};"],
                            "}",
                        ]
                    )
                elif row.type.startswith("bitfield"):
                    getter.append(f"{sunspec_variable}.raw = {value};")
                    # # below because parsesunspec only detects bitfields
                    # # if they have values
                    # if self.wrapped.enumeration_uuid is not None:
                    #     getter.append(
                    #         f'{sunspec_variable}.raw = {value};'
                    #     )
                    # else:
                    #     getter.append(
                    #         f'*((uint{row.type[-2:]}_t*) &{sunspec_variable})'
                    #         f' = {value};'
                    #     )
                else:
                    getter.append(f"{sunspec_variable} = {value};")

                setter.append("// point not implemented, do nothing")
            elif parameter.nv_format is not None:
                internal_variable = parameter.nv_format.format(meta)

                # TODO: CAMPid 075780541068182645821856068542023499
                converter = {
                    "uint32": {
                        "get": "sunspecUint32ToSSU32",
                        "set": "sunspecSSU32ToUint32",
                    },
                    "int32": {
                        # TODO: add this to embedded?
                        # 'get': 'sunspecInt32ToSSS32',
                        "set": "sunspecSSS32ToInt32",
                    },
                }.get(row.type)

                if converter is not None:
                    get_converter = converter["get"]
                    set_converter = converter["set"]

                    get_cast = ""
                    set_cast = ""
                    if parameter.nv_cast:
                        set_cast = f"(__typeof__({internal_variable})) "
                        get_type = {
                            "uint32": "uint32_t",
                        }[row.type]
                        get_cast = f"({get_type})"

                    getter.extend(
                        [
                            f"{get_converter}(",
                            [
                                f"&{sunspec_variable},",
                                f"{get_cast}{internal_variable}",
                            ],
                            ");",
                        ]
                    )
                    setter.extend(
                        [
                            f"{internal_variable} = {set_cast}{set_converter}(",
                            [
                                f"&{sunspec_variable}",
                            ],
                            ");",
                        ]
                    )
                else:
                    getter.append(
                        adjust_assignment(
                            left_hand_side=sunspec_variable,
                            right_hand_side=internal_variable,
                            sunspec_model_variable=sunspec_model_variable,
                            scale_factor=row.scale_factor,
                            internal_scale=parameter.internal_scale_factor,
                            parameter=parameter,
                            factor_operator="*",
                        )
                    )

                    setter.append(
                        adjust_assignment(
                            left_hand_side=internal_variable,
                            right_hand_side=sunspec_variable,
                            sunspec_model_variable=sunspec_model_variable,
                            scale_factor=row.scale_factor,
                            internal_scale=parameter.internal_scale_factor,
                            parameter=parameter,
                            factor_operator="/",
                        )
                    )

                # minimum_variable = parameter.nv_format.format('[Meta_Min]')
                # maximum_variable = parameter.nv_format.format('[Meta_Max]')
            elif uses_interface_item:
                # TODO: CAMPid 9685439641536675431653179671436
                parameter_uuid = str(parameter.uuid).replace("-", "_")
                item_name = f"interfaceItem_{parameter_uuid}"

                getter.extend(
                    [
                        f"{item_name}.common.sunspec.getter(",
                        [
                            f"(InterfaceItem_void *) &{item_name},",
                            f"Meta_Value",
                        ],
                        f");",
                    ]
                )
                setter.extend(
                    [
                        f"{item_name}.common.sunspec.setter(",
                        [
                            f"(InterfaceItem_void *) &{item_name},",
                            f"true,",
                            f"Meta_Value",
                        ],
                        f");",
                    ]
                )
            else:
                if getattr(parameter, "sunspec_getter", None) is not None:
                    getter.append(
                        parameter.sunspec_getter.format(
                            interface=sunspec_variable,
                        )
                    )

                if getattr(parameter, "sunspec_setter", None) is not None:
                    setter.append(
                        parameter.sunspec_setter.format(
                            interface=sunspec_variable,
                        )
                    )

            row.get = epcpm.c.format_nested_lists(getter)

            if not uses_interface_item and not self.wrapped.not_implemented:
                setter.append(f"{hand_coded_setter_function_name}();")

            if not parameter.read_only:
                row.set = epcpm.c.format_nested_lists(setter)
            else:
                row.set = None

        row.field_type = self.model_type

        if self.parameter_uuid_finder(self.wrapped.type_uuid).name == "pad":
            row.name = "Pad"
            row.description = "Force even alignment"
            row.read_write = "R"
            row.mandatory = "O"

        return row, row.size


def adjust_assignment(
    left_hand_side,
    right_hand_side,
    sunspec_model_variable,
    scale_factor,
    internal_scale,
    parameter,
    factor_operator,
):
    if scale_factor is not None:
        scale_factor_variable = f"{sunspec_model_variable}.{scale_factor}"
        # TODO: what about positive scalings?
        # factor = f'(P99_IPOW(-{scale_factor_variable}, 10))'
        # TODO: we really don't want doubles here, do we?
        # factor = f'(pow(10, -{scale_factor_variable}))'

        opposite = "" if factor_operator == "*" else "-"

        right_hand_side = (
            f"(sunspecScale({right_hand_side},"
            f" {opposite}({scale_factor_variable} + {internal_scale})))"
        )

    if parameter.nv_cast:
        right_hand_side = f"((__typeof__({left_hand_side})) {right_hand_side})"

    result = f"{left_hand_side} = {right_hand_side};"

    return result


def getter_setter_name(get_set, parameter, model_id, is_table):
    if is_table:
        table_option = "_{table_option}"
    else:
        table_option = ""

    format_string = "{get_set}SunspecModel{model_id}{table_option}_{abbreviation}"

    return format_string.format(
        get_set=get_set,
        model_id=model_id,
        abbreviation=parameter.abbreviation,
        table_option=table_option,
    )


def getter_name(parameter, model_id, is_table):
    return getter_setter_name(
        get_set="get",
        parameter=parameter,
        model_id=model_id,
        is_table=is_table,
    )


def setter_name(parameter, model_id, is_table):
    return getter_setter_name(
        get_set="set",
        parameter=parameter,
        model_id=model_id,
        is_table=is_table,
    )



import csv
import pathlib

import epcpm.project
import epcpm.smdxtosunspec
import epcpm.sunspectoxlsx


this = pathlib.Path(__file__).resolve()
here = this.parent


smdx_path = here / "sunspec"


def test_x():
    project = epcpm.project.loadp(here / "project" / "project.pmp")

    attrs_model = project.models.sunspec
    parameter_model = project.models.parameters

    enumerations = parameter_model.list_selection_roots["enumerations"]
    sunspec_types = epcpm.sunspecmodel.build_sunspec_types_enumeration()
    enumerations.append_child(sunspec_types)
    parameter_model.list_selection_roots["sunspec types"] = sunspec_types

    requested_models = [1, 17, 103, 65534]
    sunspec_models = epcpm.smdxtosunspec.import_models(
        *requested_models,
        parameter_model=parameter_model,
        paths=[smdx_path],
    )

    for sunspec_model in sunspec_models:
        attrs_model.root.append_child(sunspec_model)

    points = (
        (model, block, point)
        for model in attrs_model.root.children
        if isinstance(model, epcpm.sunspecmodel.Model)
        for block in model.children
        for point in block.children
    )

    get_set = epcpm.smdxtosunspec.import_get_set(
        smdx_path / "MODBUS_SunSpec-EPC.xlsx",
    )

    for model, block, point in points:
        parameter = attrs_model.node_from_uuid(point.parameter_uuid)
        for direction in ("get", "set"):
            key = epcpm.smdxtosunspec.GetSetKey(
                model=model.id,
                name=parameter.abbreviation,
                get_set=direction,
            )
            accessor = get_set.get(key)
            if accessor is not None:
                setattr(point, direction, accessor)

    project.filename = here / "project_with_sunspec" / "project.pmp"
    project.paths["sunspec"] = "sunspec.json"
    project.filename.parent.mkdir(parents=True, exist_ok=True)
    project.save()

    builder = epcpm.sunspectoxlsx.builders.wrap(
        wrapped=attrs_model.root,
        parameter_uuid_finder=attrs_model.node_from_uuid,
        parameter_model=project.models.parameters,
        column_filter=epcpm.sunspectoxlsx.attr_fill(epcpm.sunspectoxlsx.Fields, True),
    )

    workbook = builder.gen()

    assert workbook.sheetnames == [
        "License Agreement",
        "Summary",
        "Index",
        "1",
        "17",
        "103",
        "65534",
    ]

    workbook.save("test_sunspectoxlsx.xlsx")

    with open("test_sunspectoxlsx.csv", "w", newline="") as file:
        writer = csv.writer(file)

        for sheet in workbook.worksheets:
            for row in sheet.rows:
                writer.writerow(cell.value for cell in row)



import textwrap

import graham

import epcpm.project

import pathlib

import epyqlib.pm

reference_string = textwrap.dedent(
    """\
{
    "_type": "project",
    "paths": {
        "_type": "models",
        "parameters": "parameters.json",
        "can": "can.json",
        "sunspec": "sunspec.json",
        "staticmodbus": "staticmodbus.json"
    }
}"""
)

reference_project = epcpm.project.Project()
reference_project.paths.parameters = "parameters.json"
reference_project.paths.can = "can.json"
reference_project.paths.sunspec = "sunspec.json"
reference_project.paths.staticmodbus = "staticmodbus.json"


def test_save():
    assert reference_string == graham.dumps(reference_project, indent=4).data


def test_load():
    assert reference_project == (
        graham.schema(epcpm.project.Project).loads(reference_string).data
    )


def test_model_iterable():
    model = epcpm.project.Models()

    assert tuple(model) == ("parameters", "can", "sunspec", "staticmodbus")


def test_model_set_all():
    model = epcpm.project.Models()

    value = object()

    model.set_all(value)

    assert all(v is value for v in model.values())


def test_model_items():
    model = epcpm.project.Models()

    expected = (
        ("parameters", None),
        ("can", None),
        ("sunspec", None),
        ("staticmodbus", None),
    )

    assert tuple(model.items()) == expected


def test_model_values():
    model = epcpm.project.Models()

    for name in model:
        model[name] = name + "_"

    assert tuple(model.values()) == ("parameters_", "can_", "sunspec_", "staticmodbus_")


def test_model_getitem():
    model = epcpm.project.Models()

    values = []
    for name in model:
        values.append(model[name])

    assert values == [None, None, None, None]


def test_model_proper_selection_roots():
    project = epcpm.project.loadp(
        pathlib.Path(__file__).with_name("example_project.pmp")
    )

    expected = epyqlib.pm.parametermodel.types.list_selection_roots()

    assert set(project.models.parameters.list_selection_roots.keys()) == expected



import collections
import pathlib

import epyqlib.tests.test_attrsmodel

import epcpm.project
import epcpm.staticmodbusmodel


# See file COPYING in this source tree
__copyright__ = "Copyright 2021, EPC Power Corp."
__license__ = "GPLv2+"


this = pathlib.Path(__file__).resolve()
here = this.parent


TestAttrsModel = epyqlib.attrsmodel.build_tests(
    types=epcpm.staticmodbusmodel.types,
    root_type=epcpm.staticmodbusmodel.Root,
    columns=epcpm.staticmodbusmodel.columns,
)


# def test_model_has_header():
#     project = epcpm.project.loadp(here / "project" / "project.pmp")
#
#     parameter_model = project.models.parameters
#     enumerations = parameter_model.list_selection_roots["enumerations"]
#     staticmodbus_types = epcpm.staticmodbusmodel.build_staticmodbus_types_enumeration()
#     enumerations.append_child(staticmodbus_types)
#     parameter_model.list_selection_roots["staticmodbus types"] = staticmodbus_types
#
#     types = project.models.parameters.list_selection_roots["staticmodbus types"]
#
#     model = epcpm.staticmodbusmodel.Model()
#     parameters = model.children[0].add_data_points(
#         uint16_uuid=types.child_by_name("uint16").uuid,
#         model_id=1,
#     )
#
#     header_block = model.children[0]
#
#     assert isinstance(header_block, epcpm.staticmodbusmodel.HeaderBlock)
#
#     assert len(header_block.children) == len(parameters)
#     for point, parameter in zip(header_block.children, parameters):
#         assert point.parameter_uuid == parameter.uuid
#
#     assert parameters[0].abbreviation == "ID"
#     assert parameters[1].abbreviation == "L"
#
#     # assert header_block.offset == 0
#
#     # fixed_block = model.children[1]
#
#     # assert fixed_block.offset == 2


# TODO: CAMPid 094329054780541680163054608431067542971349
def count_types(sequence):
    counts = collections.defaultdict(int)

    for element in sequence:
        counts[type(element)] += 1

    return counts


def test_table_update_unlinked():
    expected_counts = {
        epcpm.staticmodbusmodel.TableRepeatingBlock: 8,
        epcpm.staticmodbusmodel.FunctionData: 2,
    }

    project = epcpm.project.loadp(here / "project" / "project.pmp")
    (staticmodbus_table,) = project.models.staticmodbus.root.nodes_by_attribute(
        attribute_value="First Table",
        attribute_name="name",
    )
    parameter_table = project.models.parameters.node_from_uuid(
        staticmodbus_table.parameter_table_uuid,
    )
    parameter_table.update()

    assert count_types(staticmodbus_table.children) == expected_counts

    staticmodbus_table.update()
    assert count_types(staticmodbus_table.children) == expected_counts

    print()
    print(epyqlib.utils.qt.indented_text_from_model(project.models.staticmodbus.model))

    staticmodbus_table.parameter_table_uuid = None
    staticmodbus_table.update()

    assert count_types(staticmodbus_table.children) == {}



import click
import graham
import pycparser.c_ast
import pycparser.c_generator

import epyqlib.attrsmodel
import epyqlib.pm.parametermodel

import epcpm.parameterstoc


@click.command()
@click.option("--parameters", type=click.File(), required=True)
@click.option("--declaration/--instantiation", default=True)
def cli(parameters, declaration):
    root_schema = graham.schema(epyqlib.pm.parametermodel.Root)
    root = root_schema.loads(parameters.read()).data

    model = epyqlib.attrsmodel.Model(
        root=root,
        columns=epyqlib.pm.parametermodel.columns,
    )

    builder = epcpm.parameterstoc.builders.wrap(model.root)

    generator = pycparser.c_generator.CGenerator()

    if declaration:
        items = builder.definition()
    else:
        items = builder.instantiation()

    ast = pycparser.c_ast.FileAST(items)
    s = generator.visit(ast)

    click.echo(s, nl=False)



"""
Used by parameterstoc script to build a parameters model.

It is possible that this module is either out of date or
no longer applicable for normal use.
"""
import itertools
import functools

import attr
import pycparser

import epyqlib.pm.parametermodel
import epyqlib.utils.general


builders = epyqlib.utils.general.TypeMap()


@builders(epyqlib.pm.parametermodel.Root)
@attr.s
class Root:
    wrapped = attr.ib()

    def definition(self):
        definitions = []

        for member in self.wrapped.children:
            builder = builders.wrap(member)
            definitions.extend(builder.definition())

        return definitions

    def instantiation(self):
        decls = []

        for member in self.wrapped.children:
            builder = builders.wrap(member)

            decls.append(
                Decl(
                    type=Type(
                        name=epyqlib.utils.general.spaced_to_lower_camel(
                            member.name,
                        ),
                        type=builder.type_name(),
                    )
                )
            )

        return decls


@builders(epyqlib.pm.parametermodel.Parameter)
@attr.s
class Parameter:
    wrapped = attr.ib()

    def definition(self):
        return []

    def type_name(self):
        type_name = self.wrapped.type_name

        if type_name is None:
            return "void"

        return type_name


@builders(epyqlib.pm.parametermodel.Group)
@attr.s
class Group:
    wrapped = attr.ib()

    def definition(self):
        definitions = []
        member_decls = []

        for member in self.wrapped.children:
            builder = builders.wrap(member)

            member_decls.append(
                Decl(
                    type=Type(
                        name=epyqlib.utils.general.spaced_to_lower_camel(
                            member.name,
                        ),
                        type=builder.type_name(),
                    )
                )
            )

            definitions.extend(builder.definition())

        return [
            *definitions,
            *struct(
                name=self.type_name()[:-2],
                member_decls=member_decls,
            ),
        ]

    def type_name(self):
        name = self.wrapped.type_name
        if name is None:
            name = self.wrapped.name
        return epyqlib.utils.general.spaced_to_upper_camel(name) + "_t"


@builders(epyqlib.pm.parametermodel.Array)
@attr.s
class Array:
    wrapped = attr.ib()

    def definition(self):
        builder = builders.wrap(self.wrapped.children[0])
        definitions = builder.definition()

        values = []
        if self.wrapped.named_enumerators:
            values.extend(
                enumerate(
                    epyqlib.utils.general.spaced_to_upper_camel(child.name)
                    for child in self.wrapped.children
                )
            )

        values.append((len(self.wrapped.children), "Count"))

        enum_definitions = enum(
            name=self.base_type_name(),
            enumerators=[
                (
                    "{base}_{name}".format(
                        base=self.base_type_name(),
                        name=name,
                    ),
                    value,
                )
                for value, name in values
            ],
        )

        return [
            *definitions,
            *enum_definitions,
            array_typedef(
                target=builder.type_name(),
                name=self.type_name(),
                length=pycparser.c_ast.ID(
                    enum_definitions[0].type.values.enumerators[-1].name
                ),
            ),
        ]

    def base_type_name(self):
        return epyqlib.utils.general.spaced_to_upper_camel(self.wrapped.name)

    def type_name(self):
        return self.base_type_name() + "_t"


def int_literal(value):
    return pycparser.c_ast.Constant(type="int", value=str(value))


def Type(name, type):
    return pycparser.c_ast.TypeDecl(
        declname=name,
        quals=[],
        align=None,
        type=pycparser.c_ast.IdentifierType(
            names=(type,),
        ),
    )


Decl = functools.partial(
    pycparser.c_ast.Decl,
    name=None,
    quals=[],
    align=None,
    storage=[],
    funcspec=[],
    init=None,
    bitsize=None,
)

ArrayDecl = functools.partial(
    pycparser.c_ast.ArrayDecl,
    dim_quals=[],
)

TypeDecl = functools.partial(
    pycparser.c_ast.TypeDecl,
    declname="",
    quals=[],
    align=None,
)


def typedef(target, name):
    return pycparser.c_ast.Typedef(
        name="",
        quals=[],
        storage=["typedef"],
        type=pycparser.c_ast.TypeDecl(
            declname=name,
            quals=[],
            align=None,
            type=target,
        ),
    )


def array_typedef(target, name, length):
    return pycparser.c_ast.Typedef(
        name=name,
        quals=[],
        storage=["typedef"],
        type=ArrayDecl(
            dim=length,
            type=TypeDecl(
                declname=name,
                type=pycparser.c_ast.IdentifierType(
                    names=[target],
                ),
            ),
        ),
    )


def enum(name, enumerators=()):
    enum_name = "{name}_e".format(name=name)
    typedef_name = "{name}_et".format(name=name)

    enumerators = pycparser.c_ast.EnumeratorList(
        enumerators=tuple(
            pycparser.c_ast.Enumerator(name=name, value=int_literal(value))
            for name, value in enumerators
        )
    )

    enumeration = pycparser.c_ast.Enum(name=enum_name, values=enumerators)

    declaration = Decl(type=enumeration)

    typedef_ = typedef(
        target=pycparser.c_ast.Enum(name=enum_name, values=None),
        name=typedef_name,
    )

    return declaration, typedef_


def struct(name, member_decls=()):
    struct_name = f"{name}_s"
    typedef_name = f"{name}_t"

    struct = pycparser.c_ast.Struct(
        name=struct_name,
        decls=member_decls,
    )

    decl = pycparser.c_ast.Decl(
        name=None,
        quals=[],
        align=None,
        storage=[],
        funcspec=[],
        type=struct,
        init=None,
        bitsize=None,
    )

    typedef_ = typedef(
        target=pycparser.c_ast.Struct(
            name=struct_name,
            decls=[],
        ),
        name=typedef_name,
    )

    return decl, typedef_


def array(type, name, length):
    return Decl(
        name=name,
        type=ArrayDecl(
            type=Type(name, type),
            dim=int_literal(length),
        ),
    )



import textwrap

import epcpm.c


def test_format_nested_lists():
    example = [
        "void getSUNSPEC_MODEL17_Nam(void) {",
        [
            "size_t i;",
            "UartName name = modbusHandlerGetName()",
            "",
            "for (i = 0; i < LENGTHOF(name.s); i++) {",
            [
                "sunspecInterface.model17.Nam[i] = name.s[i];",
            ],
            "}",
        ],
        "}",
    ]

    result = epcpm.c.format_nested_lists(it=example)

    assert result == textwrap.dedent(
        """\
    void getSUNSPEC_MODEL17_Nam(void) {
        size_t i;
        UartName name = modbusHandlerGetName()

        for (i = 0; i < LENGTHOF(name.s); i++) {
            sunspecInterface.model17.Nam[i] = name.s[i];
        }
    }
    """
    )



import os
import pathlib
import subprocess
import sys
import textwrap

import click
import epyqlib.pm.valuesetmodel
import lxml.etree

import epcpm.__main__
import epcpm.cli.exportdocx
import epcpm.cli.utils
import epcpm.importexport
import epcpm.importexportdialog
import epcpm.project
import epcpm.smdx


@click.group()
def main():
    """Parameter manager"""


main.add_command(epcpm.__main__._entry_point, name="gui")


@main.group(name="import")
def _import():
    """Import PM data from other formats"""
    pass


@_import.command()
@epcpm.cli.utils.project_option(required=True)
@epcpm.cli.utils.target_path_option(required=True)
def full(project, target_path):
    """Import PM data from embedded project directory"""
    project = pathlib.Path(project)

    paths = epcpm.importexportdialog.paths_from_directory(target_path)

    imported_project = epcpm.importexport.full_import(
        paths=paths,
    )

    project.parent.mkdir(exist_ok=True)
    imported_project.filename = project
    imported_project.save()


@main.group()
def export():
    """Export PM data to other formats"""
    pass


export.add_command(epcpm.cli.exportdocx.cli, name="docx")


@export.command()
@epcpm.cli.utils.project_option(required=True)
@epcpm.cli.utils.target_path_option(required=True)
@click.option("--if-stale/--assume-stale", "only_if_stale")
@click.option("--skip-sunspec/--generate-sunspec", "skip_sunspec")
@click.option(
    "--include-uuid-in-item/--exclude-uuid-from-item",
    "include_uuid_in_item",
    default=False,
)
def build(
    project,
    target_path,
    only_if_stale,
    skip_sunspec,
    include_uuid_in_item,
):
    """Export PM data to embedded project directory"""
    project = pathlib.Path(project)
    target_path = pathlib.Path(target_path)

    paths = epcpm.importexportdialog.paths_from_directory(target_path)

    if only_if_stale:
        if not epcpm.importexport.is_stale(
            project=project,
            paths=paths,
            skip_sunspec=skip_sunspec,
        ):
            click.echo(
                "Generated files appear to be up to date, skipping export",
            )

            return

        click.echo("Generated files appear to be out of date, starting export")

    loaded_project = epcpm.project.loadp(project)

    epcpm.importexport.full_export(
        project=loaded_project,
        target_directory=target_path,
        paths=paths,
        first_time=False,
        skip_sunspec=skip_sunspec,
        include_uuid_in_item=include_uuid_in_item,
    )

    click.echo()
    click.echo("done")


@main.group()
def validate():
    pass


@validate.command()
@click.option(
    "--reference",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    required=True,
)
@click.option(
    "--subject",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    required=True,
)
@click.option(
    "--schema",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
)
@click.option("--smdx-glob", default="smdx_*.xml")
def batch(reference, schema, subject, smdx_glob):
    failed = False

    reference_directory_path = pathlib.Path(reference)
    subject_directory_path = pathlib.Path(subject)

    if schema is None:
        schema = reference_directory_path / "smdx.xsd"
    else:
        schema = pathlib.Path(schema)

    paired_paths = epcpm.smdx.PairedPaths.from_directories(
        left_path=reference_directory_path,
        right_path=subject_directory_path,
        file_glob=smdx_glob,
    )

    schema = lxml.etree.fromstring(schema.read_bytes())
    schema = lxml.etree.XMLSchema(schema, attribute_defaults=True)

    spacing = "\n\n"
    present_spacing = ""

    diff_indent = "        "

    for reference_path, subject_path in sorted(paired_paths.pairs.items()):
        click.echo(present_spacing, nl=False)
        present_spacing = spacing

        click.echo(
            textwrap.dedent(
                f"""\
        Cross validating: {subject_path.name}
               reference: {reference_path}
                 subject: {subject_path}
        """
            )
        )

        reference = lxml.etree.fromstring(reference_path.read_bytes())
        subject = lxml.etree.fromstring(subject_path.read_bytes())

        result = epcpm.smdx.validate_against_reference(
            subject=subject,
            schema=schema,
            reference=reference,
        )

        if result.failed:
            failed = True

        for line in result.notes.splitlines():
            click.echo(diff_indent + line)

    for subject in sorted(paired_paths.only_right):
        click.echo(present_spacing, nl=False)
        present_spacing = spacing

        click.echo(
            textwrap.dedent(
                f"""\
        Validating: {subject.name}
           subject: {subject}
        """
            )
        )

        result = epcpm.smdx.validate_against_schema(
            subject=subject,
            schema=schema,
        )

        if result.failed:
            failed = True

        for line in result.notes.splitlines():
            click.echo(diff_indent + line)

    sys.exit(failed)


@main.command()
@epcpm.cli.utils.target_path_option(required=True)
def transition(target_path):
    """Don't use this unless you know"""
    target_path = pathlib.Path(target_path)

    click.echo("Working in: {}".format(target_path))
    value = click.prompt(
        "This will wipe out changes in the above project path, continue? ",
        prompt_suffix="",
    )

    if value != "yep":
        click.echo("Sorry, that response is not acceptable to continue.")
        return

    c_project = target_path / ".cproject"

    library_path = target_path / "embedded-library"

    original_spreadsheet = library_path / "MODBUS_SunSpec-EPC.xls"
    new_spreadsheet = original_spreadsheet.with_suffix(".xlsx")

    tables_py = library_path / "python" / "embeddedlibrary" / "tables.py"
    sunspecparser_py = library_path / "python" / "embeddedlibrary" / "sunspecparser.py"

    subprocess.run(["git", "reset", "."], check=True, cwd=library_path)
    subprocess.run(["git", "checkout", "--", "."], check=True, cwd=library_path)
    subprocess.run(["git", "clean", "-fdx"], check=True, cwd=library_path)
    subprocess.run(
        [
            "libreoffice",
            "--convert-to",
            "xlsx",
            "--outdir",
            os.fspath(library_path),
            os.fspath(original_spreadsheet),
        ],
        check=True,
        cwd=library_path,
    )
    subprocess.run(
        ["git", "rm", os.fspath(tables_py)],
        check=True,
        cwd=library_path,
    )
    subprocess.run(
        ["git", "rm", os.fspath(original_spreadsheet)],
        check=True,
        cwd=library_path,
    )
    subprocess.run(
        ["git", "add", os.fspath(new_spreadsheet)],
        check=True,
        cwd=library_path,
    )

    subprocess.run(["git", "reset", "."], check=True, cwd=target_path)
    subprocess.run(["git", "checkout", "--", "."], check=True, cwd=target_path)
    subprocess.run(
        ["git", "clean", "-fdx", "--exclude", ".venv"],
        check=True,
        cwd=target_path,
    )
    subprocess.run(
        ["python", os.fspath(target_path / "create_venv.py"), "ensure"],
        check=True,
        cwd=target_path,
    )
    subprocess.run(
        [target_path / ".venv" / "bin" / "sunspecparser", os.fspath(new_spreadsheet)],
        check=True,
        cwd=library_path,  # it expects to be in _some_ subdirectory and then ..
    )
    subprocess.run(
        ["sed", "-i", r"s/\.xls/\.xlsx/g", os.fspath(c_project)],
        check=True,
        cwd=target_path,
    )
    subprocess.run(
        ["git", "add", os.fspath(c_project)],
        check=True,
        cwd=target_path,
    )

    content = sunspecparser_py.read_text()
    with sunspecparser_py.open("w", newline="\n") as f:
        for line in content.splitlines():
            f.write(line + "\n")
            if r"""'#include "faultHandler.h"\n'""" in line:
                f.write(
                    r"""            c_file.write('#include "sunspecInterface{:>05}.h"\n'.format(model))"""
                    "\n"
                )
                f.write(r"""            c_file.write('#include "math.h"\n')""" "\n")

    subprocess.run(
        ["git", "add", os.fspath(sunspecparser_py)],
        check=True,
        cwd=library_path,
    )

    paths = epcpm.importexportdialog.paths_from_directory(target_path)
    print(paths)

    project = epcpm.importexport.full_import(
        paths=paths,
    )

    pm_directory = target_path / "interface" / "pm"
    pm_directory.mkdir(exist_ok=True)
    project.filename = pm_directory / "project.pmp"
    project.save()

    subprocess.run(
        ["git", "add", os.fspath(pm_directory)],
        check=True,
        cwd=target_path,
    )

    epcpm.importexport.full_export(
        project=project,
        paths=paths,
        first_time=True,
    )

    click.echo()
    click.echo("done")


@main.group()
def pmvs():
    pass


@pmvs.command()
@epcpm.cli.utils.project_option(required=True)
@click.option("--input", type=click.File())
@click.option("--output", type=click.Path(dir_okay=False))
def filter(project, input, output):
    """Export PM data to embedded project directory"""
    project = pathlib.Path(project)
    project = epcpm.project.loadp(project)

    value_set = epyqlib.pm.valuesetmodel.load(input)
    items = epcpm.parameterstosil.collect_items(project.models.parameters.root)
    item_uuids = {item.uuid for item in items}

    values = list(value_set.model.root.children)

    for index, value in reversed(list(enumerate(values))):
        if value.parameter_uuid not in item_uuids:
            value_set.model.root.remove_child(row=index)

    value_set.save(path=output)
